/* 📖 Bindery v2.3.0 */
(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
    typeof define === 'function' && define.amd ? define(factory) :
    (global = global || self, global.Bindery = factory());
}(this, (function () { 'use strict';

    const BINDERY_VERSION = 'v2.3.0';
    const BINDERY_CLASS_PREFIX = '📖-';
      

    function ___$insertStyle(css) {
      if (!css) {
        return;
      }
      if (typeof window === 'undefined') {
        return;
      }

      var style = document.createElement('style');

      style.setAttribute('type', 'text/css');
      style.innerHTML = css;
      document.head.appendChild(style);
      return css;
    }

    const classPrefix = BINDERY_CLASS_PREFIX;
    var SheetLayout;
    (function (SheetLayout) {
        SheetLayout["PAGES"] = "pages";
        SheetLayout["SPREADS"] = "spreads";
        SheetLayout["BOOKLET"] = "booklet";
    })(SheetLayout || (SheetLayout = {}));
    var SheetMarks;
    (function (SheetMarks) {
        SheetMarks[SheetMarks["NONE"] = 0] = "NONE";
        SheetMarks[SheetMarks["CROP"] = 1] = "CROP";
        SheetMarks[SheetMarks["BLEED"] = 2] = "BLEED";
        SheetMarks[SheetMarks["BOTH"] = 3] = "BOTH";
    })(SheetMarks || (SheetMarks = {}));
    var SheetSize;
    (function (SheetSize) {
        SheetSize["AUTO"] = "auto";
        SheetSize["AUTO_BLEED"] = "auto-bleed";
        SheetSize["AUTO_MARKS"] = "auto-marks";
        SheetSize["LETTER_PORTRAIT"] = "letter-portrait";
        SheetSize["LETTER_LANDSCAPE"] = "letter-landscape";
        SheetSize["A4_PORTRAIT"] = "a4-portrait";
        SheetSize["A4_LANDSCAPE"] = "a4-landscape";
    })(SheetSize || (SheetSize = {}));
    var ViewerMode;
    (function (ViewerMode) {
        ViewerMode["FLIPBOOK"] = "flipbook";
        ViewerMode["PREVIEW"] = "preview";
        ViewerMode["PRINT"] = "print";
    })(ViewerMode || (ViewerMode = {}));

    //
    const prefixer = (str) => {
        if (str[0] === '.') {
            return `.${classPrefix}${str.substr(1)}`;
        }
        return `${classPrefix}${str}`;
    };

    const classes = {
        showBleed: 'show-bleed',
        showCrop: 'show-crop',
        showBleedMarks: 'show-bleed-marks',
        isViewing: 'viewing',
        viewPreview: 'view-preview',
        viewPrint: 'view-print',
        viewFlip: 'view-flip',
        inProgress: 'in-progress',
        leftPage: 'left',
        rightPage: 'right',
        isOverflowing: 'is-overflowing',
        printSheet: 'print-sheet',
        sheetSpread: 'print-sheet-spread',
        sheetLeft: 'print-sheet-left',
        sheetRight: 'print-sheet-right',
        toNext: 'continues',
        fromPrev: 'continuation',
    };
    Object.keys(classes).forEach(k => {
        const key = k;
        const val = classes[key];
        classes[key] = prefixer(val);
    });
    const allModeClasses = [
        classes.viewPreview,
        classes.viewPrint,
        classes.viewFlip,
    ];
    const classForMode = (mode) => {
        switch (mode) {
            case ViewerMode.PREVIEW:
                return classes.viewPreview;
            case ViewerMode.PRINT:
                return classes.viewPrint;
            case ViewerMode.FLIPBOOK:
                return classes.viewFlip;
            default:
                throw Error(`Getting class for unknown mode: ${mode}`);
        }
    };

    const isFunc = (val) => typeof val === 'function';
    const h = (tagName, classNames, attrs, ...children) => {
        const el = document.createElement(tagName);
        if (classNames)
            el.className = classNames
                .split('.')
                .filter(txt => txt !== '')
                .map(prefixer)
                .join(' ');
        if (attrs)
            for (const k in attrs) {
                // @ts-ignore TODO replace with hyperscript anyways
                const v = attrs[k];
                // @ts-ignore TODO replace with hyperscript anyways
                if (isFunc(v))
                    el[k] = v;
                else
                    el.setAttribute(k, v);
            }
        if (children)
            el.append(...children);
        return el;
    };
    const div = (cls, ...children) => {
        return h('div', cls, {}, ...children);
    };
    const button = (cls, attrs, label) => {
        return h('button', cls, attrs, label);
    };
    const select = (cls, attrs, ...optionElements) => {
        return h('select', cls, attrs, ...optionElements);
    };
    const option = (attrs, label) => {
        return h('option', null, attrs, label);
    };

    const safeMeasure = (el, measureCallback) => {
        if (el.parentNode)
            return measureCallback();
        let measureArea = document.querySelector(prefixer('.measure-area'));
        if (!measureArea)
            measureArea = document.body.appendChild(div('.measure-area'));
        if (measureArea.firstElementChild !== el) {
            measureArea.innerHTML = '';
            measureArea.append(el);
        }
        const result = measureCallback();
        return result;
    };

    const doc = window.document;
    // Create stylesheet with id
    const addStylesheet = (id) => {
        const style = doc.createElement('style');
        style.id = id;
        doc.head.appendChild(style);
        return style;
    };
    // Fetch or create stylesheet with id
    const stylesheet = (id) => {
        var _a;
        return (_a = doc.querySelector(`#${id}`)) !== null && _a !== void 0 ? _a : addStylesheet(id);
    };
    // Parse html from text
    const parseHTML = (text, selector) => {
        const wrapper = doc.createElement('div');
        wrapper.innerHTML = text;
        return selector ? wrapper.querySelector(selector) : wrapper;
    };

    // https://github.com/moroshko/shallow-equal/blob/master/src/arrays.js
    const shallowEqual = (a, b) => {
        if (a === b)
            return true;
        if (!a || !b)
            return false;
        const len = a.length;
        if (b.length !== len) {
            return false;
        }
        for (let i = 0; i < len; i += 1) {
            if (a[i] !== b[i])
                return false;
        }
        return true;
    };

    const throttleFrame = () => {
        let wasCalled = false;
        let queued;
        const inner = (func) => {
            if (wasCalled) {
                queued = func;
                return;
            }
            wasCalled = true;
            func();
            requestAnimationFrame(() => {
                wasCalled = false;
                if (queued) {
                    const queuedFunc = queued;
                    queued = undefined;
                    inner(queuedFunc);
                }
            });
        };
        return inner;
    };
    const throttleTime = (ms) => {
        let wasCalled = false;
        let queued;
        const inner = (func) => {
            if (wasCalled) {
                queued = func;
                return;
            }
            wasCalled = true;
            func();
            setTimeout(() => {
                wasCalled = false;
                if (queued) {
                    const queuedFunc = queued;
                    queued = undefined;
                    inner(queuedFunc);
                }
            }, ms);
        };
        return inner;
    };

    const formatAsRanges = (pageNumbers) => {
        let str = '';
        let prevNum = pageNumbers[0];
        let isInARange = false;
        const addFirst = (num) => {
            str += `${num}`;
        };
        const continueRange = () => {
            isInARange = true;
        };
        const endRange = (endNum) => {
            isInARange = false;
            str += `–${endNum}`;
        };
        const addComma = (num) => {
            str += `, ${num}`;
        };
        const endAndAdd = (endNum, num) => {
            endRange(endNum);
            addComma(num);
        };
        const addLast = (num, isAdjacent) => {
            if (isAdjacent)
                endRange(num);
            else if (isInARange && !isAdjacent)
                endAndAdd(prevNum, num);
            else
                addComma(num);
        };
        pageNumbers.forEach((num, i) => {
            const isLast = i === pageNumbers.length - 1;
            const isAdjacent = num === prevNum + 1;
            if (i === 0)
                addFirst(num);
            else if (isLast)
                addLast(num, isAdjacent);
            else if (isAdjacent)
                continueRange();
            else if (isInARange && !isAdjacent)
                endAndAdd(prevNum, num);
            else
                addComma(num);
            prevNum = num;
        });
        return str;
    };

    const cssNumberRegEx = /^([+-]?[0-9]+(.?[0-9]+)?)(px|in|cm|mm|pt|pc)$/;
    const isLength = (str) => cssNumberRegEx.test(str);
    const parseLength = (str) => {
        if (!isLength(str))
            throw Error(`Cannot parse css length from "${str}"`);
        const matches = str.match(cssNumberRegEx);
        if (!matches) {
            throw Error(`Failed to parse css length from "${str}"`);
        }
        return {
            val: Number(matches[1]),
            unit: matches[3],
        };
    };

    var defaultPageSetup = {
        bleed: '12pt',
        size: { width: '4in', height: '6in' },
        margin: {
            inner: '24pt',
            outer: '24pt',
            bottom: '40pt',
            top: '48pt',
        },
    };

    const letter = Object.freeze({ width: '8.5in', height: '11in' });
    const a4 = Object.freeze({ width: '210mm', height: '297mm' });
    // Not a really reliable way to know this
    const supportsCustomPageSize = !!window.hasOwnProperty('chrome');
    class PageSetup {
        constructor(opts = {}, printOpts = {}) {
            var _a, _b, _c;
            this.size = (_a = opts.size) !== null && _a !== void 0 ? _a : defaultPageSetup.size;
            this.margin = (_b = opts.margin) !== null && _b !== void 0 ? _b : defaultPageSetup.margin;
            this.markLength = '12pt';
            this.paper = supportsCustomPageSize
                ? printOpts.paper || SheetSize.AUTO
                : SheetSize.AUTO_MARKS;
            this.bleed = (_c = printOpts.bleed) !== null && _c !== void 0 ? _c : defaultPageSetup.bleed;
            this.printTwoUp =
                !!printOpts.layout && printOpts.layout !== SheetLayout.PAGES;
        }
        get displaySize() {
            const width = this.printTwoUp ? this.spreadSize.width : this.size.width;
            const height = this.size.height;
            const bleed = this.bleed;
            return { width, height, bleed };
        }
        get sheetSize() {
            const width = this.printTwoUp ? this.spreadSize.width : this.size.width;
            const height = this.size.height;
            const doubleBleed = `2 * ${this.bleed}`;
            const doubleMarks = `${doubleBleed} + 2 * ${this.markLength}`;
            const singleMarks = `${this.bleed} + ${this.markLength}`;
            switch (this.paper) {
                case SheetSize.AUTO:
                    return { width, height };
                case SheetSize.AUTO_BLEED:
                    return {
                        width: `calc(${width} + ${this.printTwoUp ? doubleBleed : this.bleed})`,
                        height: `calc(${height} + ${doubleBleed})`,
                    };
                case SheetSize.AUTO_MARKS:
                    return {
                        width: `calc(${width} + ${this.printTwoUp ? doubleMarks : singleMarks})`,
                        height: `calc(${height} + ${doubleMarks})`,
                    };
                case SheetSize.LETTER_LANDSCAPE:
                    return { width: letter.height, height: letter.width };
                case SheetSize.LETTER_PORTRAIT:
                    return letter;
                case SheetSize.A4_PORTRAIT:
                    return a4;
                case SheetSize.A4_LANDSCAPE:
                    return { width: a4.height, height: a4.width };
                default:
                    throw Error(`Can't get dimensions for unknown paper size: ${this.paper}`);
            }
        }
        get spreadSize() {
            const w = parseLength(this.size.width);
            return {
                height: this.size.height,
                width: `${w.val * 2}${w.unit}`,
            };
        }
        updateStyleVars() {
            const page = this.size;
            const sheet = this.sheetSize;
            const cssVariables = {
                'spread-width': this.spreadSize.width,
                'page-width': page.width,
                'page-height': page.height,
                'sheet-width': sheet.width,
                'sheet-height': sheet.height,
                'margin-inner': this.margin.inner,
                'margin-outer': this.margin.outer,
                'margin-top': this.margin.top,
                'margin-bottom': this.margin.bottom,
                bleed: this.bleed,
                'mark-length': this.markLength,
            };
            const cssStr = Object.entries(cssVariables)
                .map(([k, v]) => {
                return `--bindery-${k}: ${v};`;
            })
                .join('');
            const rootRule = `:root { ${cssStr} }`;
            const pageRule = `@page { size: ${sheet.width} ${sheet.height}; }`;
            stylesheet('binderyPage').innerHTML = `${rootRule} ${pageRule}`;
        }
    }

    class Rule {
        constructor(options) {
            var _a;
            this.name = (_a = options.name) !== null && _a !== void 0 ? _a : 'Unnamed Bindery Rule';
            this.selector = '';
            Object.keys(options).forEach(key => {
                this[key] = options[key];
            });
        }
        setup() { }
    }

    const validateRuntimeOptions = (opts, validOpts) => {
        if (!validOpts)
            throw Error('Valid options not specified');
        Object.keys(opts).forEach(k => {
            var _a;
            if (!validOpts[k]) {
                const setName = (_a = validOpts.name) !== null && _a !== void 0 ? _a : 'This option';
                throw Error(`Unknown option in ${setName}: '${k}'`);
            }
            const val = opts[k];
            const type = validOpts[k];
            if (!type.check(val)) {
                const optName = validOpts.name ? `${validOpts.name}.${k}` : k;
                const valName = JSON.stringify(val);
                throw Error(`Invalid value for '${optName}': ${valName} is not a ${type.name}.`);
            }
        });
        return true;
    };

    const isObj = (val) => typeof val === 'object';
    const isFunc$1 = (val) => typeof val === 'function';
    const isBool = (val) => typeof val === 'boolean';
    const isStr = (val) => typeof val === 'string';
    const isNum = (val) => typeof val === 'number';
    const isArr = (val) => Array.isArray(val);
    const hasProp = (obj, k) => Object.prototype.hasOwnProperty.call(obj, k);
    const hasSameKeys = (opts, required) => {
        const keys = Object.keys(required).filter(k => k !== 'name');
        return !keys.some(k => !hasProp(opts, k));
    };
    const isShape = (template) => {
        return (input) => {
            return isObj(input) && validateRuntimeOptions(input, template);
        };
    };
    const isShapeExact = (template) => {
        return (input) => {
            return (isObj(input) &&
                hasSameKeys(input, template) &&
                validateRuntimeOptions(input, template));
        };
    };
    const isEnum = (cases) => {
        return (str) => {
            return cases.includes(str);
        };
    };
    const lengthChecker = {
        name: 'length (string with absolute units)',
        check: isLength,
    };
    const RuntimeTypes = {
        any: {
            name: 'any',
            check: () => true,
        },
        enum(...cases) {
            return {
                name: `(${cases.map(c => `"${c}"`).join(' | ')})`,
                check: isEnum(cases),
            };
        },
        shapeExact: (template) => ({
            name: `exactly ({${Object.keys(template).join(', ')}})`,
            check: isShapeExact(template),
        }),
        shape: (template) => ({
            name: `shape ({${Object.keys(template).join(', ')}})`,
            check: isShape(template),
        }),
        string: {
            name: 'string',
            check: isStr,
        },
        length: lengthChecker,
        number: {
            name: 'number',
            check: isNum,
        },
        bool: {
            name: 'bool',
            check: isBool,
        },
        func: {
            name: 'func',
            check: isFunc$1,
        },
        obj: {
            name: 'object',
            check: isObj,
        },
        array: {
            name: 'array',
            check: isArr,
        },
        margin: {
            name: 'margin ({ top, inner, outer, bottom })',
            check: isShapeExact({
                name: 'margin',
                top: lengthChecker,
                inner: lengthChecker,
                outer: lengthChecker,
                bottom: lengthChecker,
            }),
        },
        size: {
            name: 'size ({ width, height })',
            check: isShapeExact({
                name: 'size',
                width: lengthChecker,
                height: lengthChecker,
            }),
        },
    };

    class Counter extends Rule {
        constructor(options) {
            var _a, _b, _c;
            super(options);
            this.selector = '*';
            this.counterValue = 0;
            this.incrementEl = (_a = options.incrementEl) !== null && _a !== void 0 ? _a : '';
            this.resetEl = (_b = options.resetEl) !== null && _b !== void 0 ? _b : '';
            this.replaceEl = (_c = options.replaceEl) !== null && _c !== void 0 ? _c : '';
            validateRuntimeOptions(options, {
                name: 'Counter',
                replaceEl: RuntimeTypes.string,
                resetEl: RuntimeTypes.string,
                incrementEl: RuntimeTypes.string,
                replace: RuntimeTypes.func,
            });
        }
        setup() {
            this.counterValue = 0;
        }
        beforeAdd(el) {
            if (el.matches(this.incrementEl))
                this.counterValue += 1;
            if (el.matches(this.resetEl))
                this.counterValue = 0;
            if (el.matches(this.replaceEl))
                return this.createReplacement(el);
            return el;
        }
        createReplacement(element) {
            return this.replace(element, this.counterValue);
        }
        replace(element, counterValue) {
            element.textContent = `${counterValue}`;
            return element;
        }
    }

    class OutOfFlow extends Rule {
        constructor(options) {
            super(options);
            this.name = 'Out of Flow';
        }
        createOutOfFlowPages(elmt, book, makeNewPage) {
            throw Error('createOutOfFlowPages must be overridden');
        }
        beforeAdd(elmt) {
            // Avoid breaking inside this element. Once it's completely added,
            // it will moved onto the background layer.
            elmt.setAttribute('data-ignore-overflow', 'true');
            return elmt;
        }
        afterAdd(elmt, book, continueOnNewPage, makeNewPage) {
            this.createOutOfFlowPages(elmt, book, makeNewPage);
            // Catches cases when we didn't need to create a new page. but unclear
            if (this.continue !== 'same' || book.currentPage.hasOutOfFlowContent) {
                continueOnNewPage(true);
                if (this.continue === 'left' || this.continue === 'right') {
                    book.currentPage.setPreference(this.continue);
                }
            }
            return elmt;
        }
    }

    // Options:
    // selector: String
    class FullBleedSpread extends OutOfFlow {
        constructor(options) {
            var _a, _b;
            options.continue = (_a = options.continue) !== null && _a !== void 0 ? _a : 'same';
            options.rotate = (_b = options.rotate) !== null && _b !== void 0 ? _b : 'none';
            super(options);
            validateRuntimeOptions(options, {
                name: 'FullBleedSpread',
                selector: RuntimeTypes.string,
                continue: RuntimeTypes.enum('next', 'same', 'left', 'right'),
                rotate: RuntimeTypes.enum('none', 'clockwise', 'counterclockwise'),
            });
        }
        createOutOfFlowPages(elmt, book, makeNewPage) {
            if (!!elmt.parentNode) {
                elmt.parentNode.removeChild(elmt);
            }
            let leftPage;
            if (book.currentPage.isEmpty) {
                leftPage = book.currentPage;
            }
            else {
                leftPage = makeNewPage();
                book.addPage(leftPage);
            }
            const rightPage = makeNewPage();
            book.addPage(rightPage);
            if (this.rotate !== 'none') {
                [leftPage, rightPage].forEach(page => {
                    const rotateContainer = div(`.rotate-container.spread-size-rotated.rotate-spread-${this.rotate}`);
                    rotateContainer.append(page.background);
                    page.element.append(rotateContainer);
                });
            }
            leftPage.background.append(elmt);
            leftPage.element.classList.add(prefixer('spread'));
            leftPage.setPreference('left');
            leftPage.isOutOfFlow = this.continue === 'same';
            leftPage.avoidReorder = true;
            leftPage.hasOutOfFlowContent = true;
            rightPage.background.append(elmt.cloneNode(true));
            rightPage.element.classList.add(prefixer('spread'));
            rightPage.setPreference('right');
            rightPage.isOutOfFlow = this.continue === 'same';
            rightPage.avoidReorder = true;
            rightPage.hasOutOfFlowContent = true;
        }
    }

    // Options:
    // selector: String
    class FullBleedPage extends OutOfFlow {
        constructor(options) {
            var _a, _b;
            options.continue = (_a = options.continue) !== null && _a !== void 0 ? _a : 'same';
            options.rotate = (_b = options.rotate) !== null && _b !== void 0 ? _b : 'none';
            super(options);
            validateRuntimeOptions(options, {
                name: 'FullBleedPage',
                selector: RuntimeTypes.string,
                continue: RuntimeTypes.enum('next', 'same', 'left', 'right'),
                rotate: RuntimeTypes.enum('none', 'inward', 'outward', 'clockwise', 'counterclockwise'),
            });
        }
        createOutOfFlowPages(elmt, book, makeNewPage) {
            if (elmt.parentNode) {
                elmt.parentNode.removeChild(elmt);
            }
            let newPage;
            if (book.currentPage.isEmpty) {
                newPage = book.currentPage;
            }
            else {
                newPage = makeNewPage();
                book.addPage(newPage);
            }
            if (this.rotate !== 'none') {
                const rotateContainer = div(`.rotate-container.page-size-rotated.rotate-${this.rotate}`);
                rotateContainer.appendChild(newPage.background);
                newPage.element.appendChild(rotateContainer);
            }
            newPage.background.appendChild(elmt);
            newPage.hasOutOfFlowContent = true;
        }
    }

    // Options:
    // selector: String
    // replace: function (HTMLElement) => HTMLElement
    class Replace extends Rule {
        constructor(options) {
            super(options);
            this.name = 'Replace';
        }
        afterAdd(element, book, continueOnNewPage, makeNewPage, overflowCallback) {
            const parent = element.parentNode;
            if (!parent) {
                console.error(element);
                throw Error(`Bindery.Replace({ selector: '${this.selector}' }).afterAdd called on element that hasn't been added.`);
            }
            const defensiveClone = element.cloneNode(true);
            const replacement = this.createReplacement(book, defensiveClone);
            parent.replaceChild(replacement, element);
            if (book.currentPage.hasOverflowed()) {
                parent.replaceChild(element, replacement);
                return overflowCallback(element);
            }
            return replacement;
        }
        createReplacement(book, element) {
            return this.replace(element);
        }
        replace(element, info) {
            element.insertAdjacentHTML('beforeend', '<sup class="bindery-sup">Default Replacement</sup>');
            return element;
        }
    }

    // Options:
    // selector: String
    // replace: function (HTMLElement, number) => HTMLElement
    // render: function (Page) => HTMLElement
    class Footnote extends Replace {
        constructor(options) {
            super(options);
            validateRuntimeOptions(options, {
                name: 'Footnote',
                selector: RuntimeTypes.string,
                replace: RuntimeTypes.func,
                render: RuntimeTypes.func,
            });
        }
        afterAdd(element, book, continueOnNewPage, makeNewPage, overflowCallback) {
            const number = book.currentPage.footer.children.length + 1;
            const footnote = div('.footnote');
            const contents = this.render(element, number);
            if (contents instanceof HTMLElement)
                footnote.appendChild(contents);
            else
                footnote.innerHTML = contents;
            book.currentPage.footer.appendChild(footnote);
            return super.afterAdd(element, book, continueOnNewPage, makeNewPage, (overflowEl) => {
                book.currentPage.footer.removeChild(footnote);
                return overflowCallback(overflowEl);
            });
        }
        createReplacement(book, element) {
            const number = book.currentPage.footer.children.length;
            return this.replace(element, number);
        }
        replace(element, number) {
            element.insertAdjacentHTML('beforeend', `<sup class="bindery-sup">${number}</sup>`);
            return element;
        }
        render(element, number) {
            return `<sup>${number}</sup> Default footnote (<a href='/bindery/docs/#footnote'>Learn how to change it</a>)`;
        }
    }

    /* 📖 Regionize v0.1.4 */
    const div$1 = (cls) => {
        const el = document.createElement('div');
        el.classList.add(cls);
        return el;
    };
    class Region {
        constructor(el) {
            this.suppressErrors = false;
            this.element = el;
            this.content = div$1('region-content');
            this.content.style.padding = '0.1px';
            this.content.style.position = 'relative';
            this.element.appendChild(this.content);
            this.path = [];
        }
        setPath(newPath) {
            this.path = newPath;
            if (newPath.length > 0)
                this.content.appendChild(newPath[0]);
        }
        get currentElement() {
            const len = this.path.length;
            if (len > 0)
                return this.path[len - 1];
            return this.content;
        }
        get isEmpty() {
            const el = this.content;
            if (el.textContent === null)
                return true;
            return el.textContent.trim() === '' && el.offsetHeight < 1;
        }
        get isReasonableSize() {
            const box = this.element.getBoundingClientRect();
            return (box.height > 100) && (box.width > 100); // TODO: Number is arbitrary
        }
        overflowAmount() {
            const contentH = this.content.offsetHeight;
            const boxH = this.element.offsetHeight;
            if (boxH === 0)
                throw Error('Regionizer: Trying to flow into an element with zero height.');
            return contentH - boxH;
        }
        hasOverflowed() {
            return this.overflowAmount() > -5;
        }
    }

    const isTextNode = (node) => {
        return node.nodeType === Node.TEXT_NODE;
    };
    const isElement = (node) => {
        return node.nodeType === Node.ELEMENT_NODE;
    };
    const isScript = (node) => {
        return node.tagName === 'SCRIPT';
    };
    const isImage = (node) => {
        return node.tagName === 'IMG';
    };
    const isUnloadedImage = (node) => {
        return isImage(node) && !(node).naturalWidth;
    };
    const isContentElement = (node) => {
        return isElement(node) && !isScript(node);
    };

    const MAX_TIME = 30; // ms
    const rAF = () => new Promise((resolve) => {
        requestAnimationFrame(t => resolve(t));
    });
    let lastYieldTime = 0;
    const shouldYield = () => {
        const timeSinceYield = performance.now() - lastYieldTime;
        return timeSinceYield > MAX_TIME;
    };
    const yieldIfNecessary = async () => {
        if (shouldYield())
            lastYieldTime = await rAF();
    };

    const overflowAttr = 'data-ignore-overflow';
    // Walk up the tree to see if we are within
    // an overflow-ignoring node
    const ignoreOverflow = (element) => {
        if (element.hasAttribute(overflowAttr))
            return true;
        if (element.parentElement)
            return ignoreOverflow(element.parentElement);
        return false;
    };

    const SPACE = ' ';
    const nextWordEnd = (text, startIndex) => {
        let newIndex = startIndex + 1;
        while (newIndex < text.length && text.charAt(newIndex) !== SPACE) {
            newIndex += 1;
        }
        return newIndex;
    };
    const previousWordEnd = (text, startIndex) => {
        let newIndex = startIndex;
        if (text.charAt(newIndex) === SPACE) {
            newIndex -= 1;
        }
        while (text.charAt(newIndex) !== SPACE && newIndex > 0) {
            newIndex -= 1;
        }
        return newIndex;
    };

    const createTextNode = (document.createTextNode).bind(document);
    // Try adding a text node in one go.
    // Returns true if all the text fits, false if none fits.
    const addInOneGo = async (textNode, container, hasOverflowed) => {
        container.appendChild(textNode);
        const success = !hasOverflowed();
        if (!success)
            container.removeChild(textNode);
        await yieldIfNecessary();
        return { completed: success };
    };
    // Incrementally add words to the container until it just barely doesnt
    // overflow. Returns a remainder textNode for remaining text.
    const fillWordsUntilOverflow = async (textNode, container, hasOverflowed) => {
        const originalText = textNode.nodeValue || '';
        container.appendChild(textNode);
        if (!hasOverflowed() || ignoreOverflow(container)) {
            // The whole thing fits
            return { completed: true };
        }
        // Clear the node
        let proposedEnd = 0;
        textNode.nodeValue = originalText.substr(0, proposedEnd);
        while (!hasOverflowed() && proposedEnd < originalText.length) {
            // Reveal the next word
            proposedEnd = nextWordEnd(originalText, proposedEnd);
            if (proposedEnd < originalText.length) {
                textNode.nodeValue = originalText.substr(0, proposedEnd);
                await yieldIfNecessary();
            }
        }
        // Back out to word boundary
        const wordEnd = previousWordEnd(originalText, proposedEnd);
        if (wordEnd < 1) {
            // We didn't even add a complete word, don't add node
            textNode.nodeValue = originalText;
            container.removeChild(textNode);
            return { completed: false };
        }
        // trim text to word
        const fittingText = originalText.substr(0, wordEnd);
        const overflowingText = originalText.substr(wordEnd);
        textNode.nodeValue = fittingText;
        // Create a new text node for the next flow box
        return {
            completed: true,
            remainder: createTextNode(overflowingText)
        };
    };
    // Fills text across multiple elements by requesting a continuation
    // once the current element overflows
    const fillWords = async (textNode, container, getNextContainer, hasOverflowed) => {
        const textLayout = await fillWordsUntilOverflow(textNode, container, hasOverflowed);
        if (textLayout.remainder) {
            const nextContainer = getNextContainer();
            return fillWords(textLayout.remainder, nextContainer, getNextContainer, hasOverflowed);
        }
        return textLayout;
    };

    // Shifts this element to the next page. If any of its
    // ancestors cannot be split across page, it will
    // step up the tree to find the first ancestor
    // that can be split, and move all of that descendants
    // to the next page.
    const tryInNextRegion = (region, makeNextRegion, canSplit) => {
        if (region.path.length <= 1) {
            throw Error('Regionize: Attempting to move the top-level element');
        }
        const startLength = region.path.length;
        // So this node won't get cloned. TODO: this is unclear
        const elementToMove = region.path.pop();
        // find the nearest splittable parent
        let nearestMoveableElement = elementToMove;
        const pathToRestore = [];
        while (region.path.length > 1 && !canSplit(region.currentElement)) {
            nearestMoveableElement = region.path.pop();
            pathToRestore.unshift(nearestMoveableElement);
        }
        // Once a node is moved to a new page, it should no longer trigger another
        // move. otherwise tall elements will endlessly get shifted to the next page
        nearestMoveableElement.setAttribute('data-bindery-did-move', 'true');
        const parent = nearestMoveableElement.parentNode;
        parent.removeChild(nearestMoveableElement);
        // If the nearest ancestor would be empty without this node,
        // move it to the next page too.
        if (region.path.length > 1 && region.currentElement.textContent.trim() === '') {
            parent.appendChild(nearestMoveableElement);
            nearestMoveableElement = region.path.pop();
            pathToRestore.unshift(nearestMoveableElement);
            nearestMoveableElement.parentNode.removeChild(nearestMoveableElement);
        }
        let nextRegion;
        if (!region.isEmpty) {
            if (region.hasOverflowed()) {
                // Recovery failed, maybe the box contains a large
                // unsplittable element.
                region.suppressErrors = true;
            }
            nextRegion = makeNextRegion();
        }
        else {
            // If the page is empty when this node is removed,
            // then it won't help to move it to the next page.
            // Instead continue here until the node is done.
            nextRegion = region;
        }
        // append moved node as first in new page
        nextRegion.currentElement.appendChild(nearestMoveableElement);
        // restore subpath
        pathToRestore.forEach(r => nextRegion.path.push(r));
        nextRegion.path.push(elementToMove);
        if (startLength !== nextRegion.path.length) {
            throw Error('Regionize: Restored path depth does not match original path depth');
        }
    };

    const clonePath = (oldPath, applyRules) => {
        const newPath = [];
        const deepClone = (el) => {
            const clone = el.cloneNode(true); // deep clone, could be th > h3 > span;
            applyRules(el, clone);
            return clone;
        };
        for (let i = oldPath.length - 1; i >= 0; i -= 1) {
            const original = oldPath[i];
            const clone = original.cloneNode(false); // shallow
            const nextChild = oldPath[i + 1];
            clone.innerHTML = '';
            applyRules(original, clone, nextChild, deepClone);
            if (i < oldPath.length - 1)
                clone.appendChild(newPath[i + 1]);
            newPath[i] = clone;
        }
        return newPath;
    };

    // Polls every 10ms for image.naturalWidth
    // or an error event.
    //
    // Note: Doesn't ever reject, since missing images
    // shouldn't prevent layout from resolving
    const wait10 = () => new Promise((resolve) => {
        setTimeout(() => { resolve(); }, 10);
    });
    const ensureImageLoaded = async (image) => {
        const imgStart = performance.now();
        let failed = false;
        image.addEventListener('error', () => { failed = true; });
        image.src = image.src; // re-trigger error if already failed
        while (!image.naturalWidth && !failed) {
            await wait10();
        }
        return performance.now() - imgStart;
    };

    const preserveNumbering = (original, clone, nextChild) => {
        // restart numbering
        let prevStart = 1;
        if (original.hasAttribute('start')) {
            // the OL is also a continuation
            prevStart = parseInt(original.getAttribute('start') || "", 10);
        }
        if (nextChild && nextChild.tagName === 'LI') {
            // the first list item is a continuation
            prevStart -= 1;
        }
        const prevCount = original.children.length;
        const newStart = prevStart + prevCount;
        clone.setAttribute('start', `${newStart}`);
    };

    const preserveTableColumns = (original, clone, nextChild, deepClone) => {
        const columns = [...original.children];
        const currentIndex = columns.indexOf(nextChild);
        for (let i = 0; i < currentIndex; i += 1) {
            const origCol = columns[i];
            if (origCol) {
                const clonedCol = deepClone(origCol);
                clone.appendChild(clonedCol);
            }
        }
    };

    const noop = () => { };
    const always = () => true;
    const never = () => false;
    // flow content through FlowBoxes.
    // the caller is responsible for managing
    // and creating regions.
    const flowIntoRegions = async (opts) => {
        const content = opts.content;
        const createRegion = opts.createRegion;
        if (!content)
            throw Error('content not specified');
        if (!createRegion)
            throw Error('createRegion not specified');
        // optional
        const applySplit = opts.applySplit || noop;
        const canSplit = opts.canSplit || always;
        const beforeAdd = opts.beforeAdd || noop;
        const afterAdd = opts.afterAdd || noop;
        const didWaitFor = opts.didWaitFor || noop;
        const shouldTraverse = opts.shouldTraverse || never;
        // currentRegion should hold the only state that persists during traversal.
        let currentRegion = createRegion();
        const hasOverflowed = () => currentRegion.hasOverflowed();
        const canSplitCurrent = () => canSplit(currentRegion.currentElement);
        const ignoreCurrentOverflow = () => ignoreOverflow(currentRegion.currentElement);
        const splitRules = (original, clone, nextChild, deepClone) => {
            if (original.tagName === 'OL') {
                preserveNumbering(original, clone, nextChild);
            }
            if (original.tagName === 'TR' && nextChild && deepClone) {
                preserveTableColumns(original, clone, nextChild, deepClone);
            }
            applySplit(original, clone, nextChild, deepClone);
        };
        const continueInNextRegion = () => {
            const prevRegion = currentRegion;
            currentRegion = createRegion();
            const newPath = clonePath(prevRegion.path, splitRules);
            currentRegion.setPath(newPath);
            return currentRegion;
        };
        const continuedParent = () => {
            continueInNextRegion();
            return currentRegion.currentElement;
        };
        const addText = async (textNode, isSplittable) => {
            const el = currentRegion.currentElement;
            let textLayout;
            if (isSplittable) {
                // Add the text word by word, adding pages as needed
                textLayout = await fillWords(textNode, el, continuedParent, hasOverflowed);
                if (!textLayout.completed && currentRegion.path.length > 1) {
                    tryInNextRegion(currentRegion, continueInNextRegion, canSplit);
                    textLayout = await fillWords(textNode, el, continuedParent, hasOverflowed);
                }
            }
            else {
                // Add the text as a block, trying a new page if needed
                textLayout = await addInOneGo(textNode, currentRegion.currentElement, hasOverflowed);
                if (!textLayout.completed && !ignoreCurrentOverflow()) {
                    tryInNextRegion(currentRegion, continueInNextRegion, canSplit);
                    textLayout = await addInOneGo(textNode, currentRegion.currentElement, hasOverflowed);
                }
            }
            // Something went wrong. Insert the text anyways, ignoring overflow,
            // and move onto the next region.
            if (!textLayout.completed) {
                currentRegion.currentElement.appendChild(textNode);
                if (!ignoreCurrentOverflow() && canSplitCurrent()) {
                    currentRegion.suppressErrors = true;
                    continueInNextRegion();
                }
            }
        };
        const shouldTraverseChildren = (element) => {
            if (hasOverflowed())
                return true;
            if (element.querySelector('img'))
                return true;
            if (shouldTraverse(element))
                return true;
            return false;
        };
        const addElement = async (element) => {
            // Ensure images are loaded before testing for overflow
            if (isUnloadedImage(element)) {
                const waitTime = await ensureImageLoaded(element);
                didWaitFor(waitTime);
            }
            // Transforms before adding
            beforeAdd(element, continueInNextRegion);
            // Append element and push onto the the stack
            currentRegion.currentElement.appendChild(element);
            currentRegion.path.push(element);
            if (shouldTraverseChildren(element)) {
                // Only if the region overflowed, the content contains
                // an image, or the caller has requested a custom traversal.
                await clearAndAddChildren(element);
            }
            // We're done: Pop off the stack and do any cleanup
            const addedElement = currentRegion.path.pop();
            afterAdd(addedElement, continueInNextRegion);
        };
        const clearAndAddChildren = async (element) => {
            const childNodes = [...element.childNodes];
            element.innerHTML = '';
            if (hasOverflowed() && !ignoreCurrentOverflow() && canSplitCurrent()) {
                // Overflows when empty
                tryInNextRegion(currentRegion, continueInNextRegion, canSplit);
            }
            const shouldSplit = canSplit(element) && !ignoreOverflow(element);
            for (const child of childNodes) {
                if (isTextNode(child)) {
                    await addText(child, shouldSplit);
                }
                else if (isContentElement(child)) {
                    await addElement(child);
                }
            }
        };
        return addElement(content);
    };

    class HierarchyToHeadingAdapter {
        constructor(getter) {
            // console.warn('Deprecated');
            this.getHierarchy = getter;
        }
        textFor(sel) {
            var _a, _b;
            return (_b = (_a = this.getHierarchy()) === null || _a === void 0 ? void 0 : _a.find(entry => (entry === null || entry === void 0 ? void 0 : entry.selector) === sel)) === null || _b === void 0 ? void 0 : _b.text;
        }
        get h1() {
            return this.textFor('h1');
        }
        get h2() {
            return this.textFor('h2');
        }
        get h3() {
            return this.textFor('h3');
        }
        get h4() {
            return this.textFor('h4');
        }
        get h5() {
            return this.textFor('h5');
        }
        get h6() {
            return this.textFor('h6');
        }
    }
    class Page {
        constructor() {
            this.hierarchy = [];
            this.suppress = false;
            this.hasOutOfFlowContent = false;
            this.alwaysLeft = false;
            this.alwaysRight = false;
            this.isOutOfFlow = false; // used by spreads
            this.avoidReorder = false; // used by 2-page spreads
            this.flow = new Region(div('flow-box'));
            this.footer = div('footer');
            this.background = div('page-background');
            this.element = div('page', this.background, this.flow.element, this.footer);
            this.heading = new HierarchyToHeadingAdapter(() => this.hierarchy);
        }
        static isSizeValid() {
            const testPage = new Page();
            return safeMeasure(testPage.element, () => testPage.flow.isReasonableSize);
        }
        setLeftRight(dir) {
            this.side = dir;
            this.element.classList.toggle(classes.leftPage, this.isLeft);
            this.element.classList.toggle(classes.rightPage, !this.isLeft);
        }
        get isLeft() {
            return this.side === 'left';
        }
        get isRight() {
            return this.side === 'right';
        }
        setPreference(dir) {
            const preferLeft = dir === 'left';
            this.alwaysLeft = preferLeft;
            this.alwaysRight = !preferLeft;
        }
        get suppressErrors() {
            var _a;
            return (_a = this.suppress) !== null && _a !== void 0 ? _a : false;
        }
        set suppressErrors(newVal) {
            this.suppress = newVal;
            this.element.classList.toggle(classes.isOverflowing, newVal);
        }
        get isEmpty() {
            return !this.hasOutOfFlowContent && this.flow.isEmpty;
        }
        validate() {
            if (!this.hasOverflowed())
                return;
            const suspect = this.flow.currentElement;
            if (suspect) {
                console.warn('Bindery: Content overflows, probably due to a style set on:', suspect);
                if (suspect.parentNode) {
                    suspect.parentNode.removeChild(suspect);
                }
            }
            else {
                console.warn('Bindery: Content overflows.');
            }
        }
        validateEnd(allowOverflow) {
            if (!this.hasOverflowed())
                return;
            console.warn(`Bindery: Page ~${this.number} is overflowing`, this.element);
            if (!this.suppressErrors && !this.flow.suppressErrors && !allowOverflow) {
                throw Error('Bindery: Moved to new page when last one is still overflowing');
            }
        }
        hasOverflowed() {
            return safeMeasure(this.element, () => this.flow.hasOverflowed());
        }
    }

    const indexOfNextReorderablePage = (pages, startIndex) => {
        for (let i = startIndex; i < pages.length; i += 1) {
            const pg = pages[i];
            if (!pg.isOutOfFlow && !pg.avoidReorder)
                return i;
        }
        return null;
    };
    // Given an array of pages with alwaysLeft, alwaysRight, and isOutOfFlow
    // properties, orders them so that alwaysLeft and alwaysRight are true.
    const orderPages = (pages, makeNewPage) => {
        const orderedPages = pages.slice();
        for (let i = 0; i < orderedPages.length; i += 1) {
            const page = orderedPages[i];
            const isLeft = i % 2 !== 0;
            if ((isLeft && page.alwaysRight) || (!isLeft && page.alwaysLeft)) {
                if (page.isOutOfFlow) {
                    // If the page is 'out of flow', we'd prefer not to add a blank page.
                    // Instead it floats backwards in the book, pulling the next
                    // in-flow page forward. If several 'out of flow' pages
                    // are next to each other, they will remain in order, all being pushed
                    // backward together.
                    const indexToSwap = indexOfNextReorderablePage(orderedPages, i + 1);
                    if (!indexToSwap) {
                        // No larger index to swap with, perhaps because
                        // we are optimistically rendering before the book is done
                        break;
                    }
                    const pageToMoveUp = orderedPages[indexToSwap];
                    orderedPages.splice(indexToSwap, 1); // remove pg
                    orderedPages.splice(i, 0, pageToMoveUp); // insert pg
                }
                else {
                    // If the page is 'in flow', order must be respected, so extra blank pages
                    // are inserted.
                    orderedPages.splice(i, 0, makeNewPage());
                }
            }
        }
        return orderedPages;
    };

    const MAXIMUM_PAGE_LIMIT = 2000;
    class Book {
        constructor() {
            this.rawPages = [];
            this.orderedPages = [];
        }
        addPage(newPage) {
            this.rawPages.push(newPage);
            this.updatePageOrder();
        }
        get pageCount() {
            return this.orderedPages.length;
        }
        get pages() {
            return this.orderedPages;
        }
        updatePageOrder() {
            this.orderedPages = orderPages(this.rawPages, () => new Page());
        }
        validate() {
            if (this.pageCount > MAXIMUM_PAGE_LIMIT) {
                throw Error('Bindery: Maximum page count exceeded. Suspected runaway layout.');
            }
        }
    }

    const annotatePagesNumbers = (pages, offset) => {
        {
            pages.forEach((page, i) => {
                page.number = offset + i + 1;
                page.setLeftRight(i % 2 === 0 ? 'right' : 'left');
            });
        }
    };
    const annotatePagesHierarchy = (pages, headerSelectorHierarchy) => {
        // ———
        // RUNNING HEADERS
        // Sections to annotate with.
        // This should be a hierarchical list of selectors.
        // Every time one is selected, it annotates all following pages
        // and clears any subselectors.
        let currentHierarchy = [];
        pages.forEach(page => {
            const pageHierarchy = [];
            headerSelectorHierarchy.forEach((selector, i) => {
                var _a;
                const element = page.element.querySelector(selector);
                // A new header level starts on this page
                if (element) {
                    currentHierarchy[i] = {
                        selector: selector,
                        text: (_a = element.textContent) !== null && _a !== void 0 ? _a : '',
                        el: element,
                    };
                    // Clear any lower headers in the hierarchy
                    currentHierarchy = currentHierarchy.slice(0, i + 1);
                    // headerSelectorHierarchy.forEach((lowerSelector, j) => {
                    //   if (j > i) {
                    //     currentHeaders[j] = { selector: lowerSelector, text: '', el: undefined };
                    //   }
                    // });
                }
                // Always decorate this page with current header state.
                if (currentHierarchy[i]) {
                    pageHierarchy[i] = currentHierarchy[i];
                }
            });
            page.hierarchy = pageHierarchy;
        });
    };
    const annotatePages = (pages, offset) => {
        annotatePagesNumbers(pages, offset);
        annotatePagesHierarchy(pages, ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']);
    };

    const pageNumbersForTest = (pages, test) => {
        return pages
            .filter(pg => !!pg.number)
            .filter(pg => test(pg.element))
            .map(pg => pg.number);
    };

    // Compatible with ids that start with numbers
    const startsWithNumber = (sel) => {
        return sel.length > 2 && sel[0] === '#' && /^\d+$/.test(sel[1]);
    };
    const safeIDSel = (sel) => {
        return startsWithNumber(sel) ? `[id="${sel.replace('#', '')}"]` : sel;
    };
    // Options:
    // selector: String
    // replace: function (HTMLElement, number) => HTMLElement
    class PageReference extends Replace {
        constructor(options) {
            super(options);
            validateRuntimeOptions(options, {
                name: 'PageReference',
                selector: RuntimeTypes.string,
                replace: RuntimeTypes.func,
                createTest: RuntimeTypes.func,
            });
            this.references = [];
            const throttle = throttleTime(10);
            this.throttledUpdate = book => {
                throttle(() => this.updatePageReferences(book.pages));
            };
        }
        eachPage(page, book) {
            this.throttledUpdate(book);
        }
        afterAdd(elmt, book) {
            const test = this.createTest(elmt);
            if (!test)
                return elmt;
            const ref = this.createReference(book, test, elmt);
            return ref.element;
        }
        createReference(book, test, elmt) {
            const ref = {
                test,
                template: elmt,
                element: elmt,
                value: null,
            };
            this.references.push(ref);
            const currentResults = pageNumbersForTest(book.pages, test);
            this.render(ref, currentResults); // Replace element immediately, to make sure it'll fit
            return ref;
        }
        render(ref, newValue) {
            if (!newValue || shallowEqual(ref.value, newValue))
                return;
            if (!Array.isArray(newValue))
                throw Error('Page search returned unexpected result');
            const isResolved = newValue.length > 0;
            const pageRanges = isResolved ? formatAsRanges(newValue) : '⌧';
            const template = ref.template.cloneNode(true);
            const newRender = this.replace(template, pageRanges);
            if (!isResolved)
                newRender.classList.add(prefixer('placeholder-num'));
            ref.element.parentNode.replaceChild(newRender, ref.element);
            ref.element = newRender;
            ref.value = newValue;
        }
        createTest(element) {
            const href = element.getAttribute('href');
            if (!href)
                return null;
            const selector = safeIDSel(href);
            return (el) => {
                return !!el.querySelector(selector);
            };
        }
        updatePageReferences(pages) {
            // querySelector first, then rerender
            const results = this.references.map(ref => {
                return { ref, data: pageNumbersForTest(pages, ref.test) };
            });
            results.forEach(({ ref, data }) => this.render(ref, data));
        }
        replace(template, number) {
            template.insertAdjacentHTML('beforeend', `, <span>${number}</span>`);
            return template;
        }
    }

    class PageBreak extends Rule {
        constructor(options) {
            var _a, _b;
            super(options);
            this.continue = (_a = options.continue) !== null && _a !== void 0 ? _a : 'next';
            this.position = (_b = options.position) !== null && _b !== void 0 ? _b : 'before';
            validateRuntimeOptions(options, {
                name: 'PageBreak',
                selector: RuntimeTypes.string,
                continue: RuntimeTypes.enum('next', 'left', 'right'),
                position: RuntimeTypes.enum('before', 'after', 'both', 'avoid'),
            });
        }
        get avoidSplit() {
            return this.position === 'avoid';
        }
        beforeAdd(elmt, book, continueOnNewPage) {
            if (this.position === 'before' || this.position === 'both') {
                if (!book.currentPage.isEmpty) {
                    continueOnNewPage();
                }
                if (this.continue !== 'next' && this.continue !== 'same') {
                    book.currentPage.setPreference(this.continue);
                }
            }
            return elmt;
        }
        afterAdd(elmt, book, continueOnNewPage) {
            if (this.position === 'after' || this.position === 'both') {
                const newPage = continueOnNewPage(true);
                if (this.continue !== 'next') {
                    newPage.setPreference(this.continue);
                }
            }
            return elmt;
        }
    }

    // Options:
    // selector: String
    // render: function (Page) => HTMLElement
    // TODO selectorHierarchy: [ String ], ie [ 'h1', 'h2', 'h3.chapter' ]
    class RunningHeader extends Rule {
        constructor(options = {}) {
            super(options);
            validateRuntimeOptions(options, {
                name: 'RunningHeader',
                render: RuntimeTypes.func,
            });
        }
        eachPage(page) {
            if (!page.runningHeader) {
                const elmt = div('.running-header');
                page.element.appendChild(elmt);
                page.runningHeader = elmt;
            }
            page.runningHeader.innerHTML = this.render(page);
        }
        render(page) {
            return `${page.number}`;
        }
    }

    class Split extends Rule {
        constructor(options) {
            super(options);
            this.toNext = options.toNext;
            this.fromPrevious = options.fromPrevious;
            validateRuntimeOptions(options, {
                name: 'Split',
                selector: RuntimeTypes.string,
                toNext: RuntimeTypes.string,
                fromPrevious: RuntimeTypes.string,
                didSplit: RuntimeTypes.func,
            });
        }
        didSplit(original, clone) {
            if (this.toNext)
                original.classList.add(this.toNext);
            if (this.fromPrevious)
                clone.classList.add(this.fromPrevious);
        }
    }

    var rules = {
        Rule,
        Split(options) {
            return new Split(options);
        },
        Counter(options) {
            return new Counter(options);
        },
        FullBleedPage(options) {
            return new FullBleedPage(options);
        },
        Footnote(options) {
            return new Footnote(options);
        },
        RunningHeader(options) {
            return new RunningHeader(options);
        },
        Replace(options) {
            return new Replace(options);
        },
        FullBleedSpread(options) {
            return new FullBleedSpread(options);
        },
        PageBreak(options) {
            return new PageBreak(options);
        },
        PageReference(options) {
            return new PageReference(options);
        },
        createRule(options) {
            return new Rule(options);
        },
    };

    const { PageBreak: PageBreak$1, PageReference: PageReference$1, Footnote: Footnote$1, FullBleedPage: FullBleedPage$1, FullBleedSpread: FullBleedSpread$1, } = rules;
    const replacer = (element, number) => {
        element.textContent = `${number}`;
        return element;
    };
    var attributeRules = [
        PageBreak$1({ selector: '[book-page-break="both"]', position: 'both' }),
        PageBreak$1({ selector: '[book-page-break="avoid"]', position: 'avoid' }),
        PageBreak$1({
            selector: '[book-page-break="after"][book-page-continue="right"]',
            position: 'after',
            continue: 'right',
        }),
        PageBreak$1({
            selector: '[book-page-break="after"][book-page-continue="left"]',
            position: 'after',
            continue: 'left',
        }),
        PageBreak$1({
            selector: '[book-page-break="after"][book-page-continue="next"]',
            position: 'after',
            continue: 'next',
        }),
        PageBreak$1({
            selector: '[book-page-break="before"][book-page-continue="right"]',
            position: 'before',
            continue: 'right',
        }),
        PageBreak$1({
            selector: '[book-page-break="before"][book-page-continue="left"]',
            position: 'before',
            continue: 'left',
        }),
        PageBreak$1({
            selector: '[book-page-break="before"][book-page-continue="next"]',
            position: 'before',
            continue: 'next',
        }),
        FullBleedPage$1({ selector: '[book-full-bleed="page"]' }),
        FullBleedSpread$1({ selector: '[book-full-bleed="spread"]' }),
        Footnote$1({
            selector: '[book-footnote-text]',
            render: (element, number) => {
                const txt = element.getAttribute('book-footnote-text');
                return `<i>${number}</i>${txt}`;
            },
        }),
        PageReference$1({
            selector: '[book-pages-with-text]',
            replace: replacer,
            createTest: (element) => {
                var _a;
                const text = (_a = element.getAttribute('book-pages-with-text')) !== null && _a !== void 0 ? _a : '';
                const term = text.toLowerCase().trim();
                return (pageElement) => {
                    const pageText = pageElement.textContent || '';
                    return pageText.toLowerCase().includes(term);
                };
            },
        }),
        PageReference$1({
            selector: '[book-pages-with-selector]',
            replace: replacer,
            createTest: (element) => {
                var _a;
                const txt = (_a = element.getAttribute('book-pages-with-selector')) !== null && _a !== void 0 ? _a : '';
                const selector = txt.trim();
                return (pageElement) => {
                    return pageElement.querySelector(selector);
                };
            },
        }),
        PageReference$1({
            selector: '[book-pages-with]',
            replace: replacer,
            createTest: (element) => {
                var _a;
                const text = (_a = element.textContent) !== null && _a !== void 0 ? _a : '';
                const term = text.toLowerCase().trim();
                return (pageElement) => {
                    var _a;
                    const pageText = (_a = pageElement.textContent) !== null && _a !== void 0 ? _a : '';
                    return pageText.toLowerCase().includes(term);
                };
            },
        }),
    ];

    const isSpread = (rule) => rule instanceof FullBleedSpread;
    const isPage = (rule) => rule instanceof FullBleedPage;
    const isBreak = (rule) => rule instanceof PageBreak;
    const isFullPageRule = (rule) => isSpread(rule) || isPage(rule) || isBreak(rule);
    const dedupe = (inputRules) => {
        const conflictRules = inputRules.filter(isFullPageRule);
        const output = inputRules.filter(rule => !conflictRules.includes(rule));
        const firstSpreadRule = conflictRules.find(isSpread);
        const firstPageRule = conflictRules.find(isPage);
        // Only apply one fullpage or fullspread
        if (firstSpreadRule)
            output.push(firstSpreadRule);
        else if (firstPageRule)
            output.push(firstPageRule);
        else
            output.push(...conflictRules); // but multiple pagebreaks are ok
        return output;
    };

    // TODO:
    // While this does catch overflows, it is pretty hacky to move the entire node to the next page.
    // - 1. there is no guarentee it will fit on the new page
    // - 2. if it had childNodes, those side effects will not be undone,
    // which means footnotes will get left on previous page.
    // - 3. if it is a large paragraph, it will leave a large gap. the
    // ideal approach would be to only need to invalidate the last line of text.
    const recoverFromRule = (el, book, nextRegion) => {
        let removed = el;
        const parent = el.parentNode;
        if (!parent) {
            throw Error("Can't recoverFromRule when element is unparented");
        }
        parent.removeChild(removed);
        let popped;
        if (book.currentPage.hasOverflowed()) {
            parent.appendChild(el);
            removed = parent;
            if (!removed.parentNode) {
                throw Error("Can't recoverFromRule when element is unparented");
            }
            removed.parentNode.removeChild(removed);
            popped = book.currentPage.flow.path.pop();
            if (book.currentPage.hasOverflowed()) {
                console.error('Trying again didnt fix it');
            }
        }
        const newRegion = nextRegion();
        newRegion.currentElement.appendChild(removed);
        if (popped)
            newRegion.path.push(popped);
    };

    const giveUp = (rule, el) => {
        console.warn(`Couldn't apply ${rule.name}, caused overflows twice when adding: `, el);
    };
    function isPageRule(rule) {
        return rule.eachPage;
    }
    function isBeforeAddRule(rule) {
        return !!rule.selector && rule.beforeAdd;
    }
    function isAfterAddRule(rule) {
        return !!rule.selector && rule.afterAdd;
    }
    function isOffsetRule(rule) {
        return rule.pageNumberOffset;
    }
    function isDidSplitRule(rule) {
        return !!rule.selector && rule.didSplit;
    }
    function isAvoidSplitRule(rule) {
        return !!rule.selector && rule.avoidSplit;
    }
    class RuleSet {
        constructor(rules) {
            var _a;
            const offsetRule = rules.find(isOffsetRule);
            this.pageNumberOffset = (_a = offsetRule === null || offsetRule === void 0 ? void 0 : offsetRule.pageNumberOffset) !== null && _a !== void 0 ? _a : 0;
            // Rules for pages
            this.pageRules = rules.filter(isPageRule);
            // Rules for elements
            this.beforeAddRules = rules.filter(isBeforeAddRule);
            this.afterAddRules = rules.filter(isAfterAddRule);
            // Rules for layout
            this.selectorsNotToSplit = rules
                .filter(isAvoidSplitRule)
                .map(r => r.selector);
            this.didSplitRules = rules.filter(isDidSplitRule);
            // setup
            rules.filter(r => r.setup).forEach(r => r.setup());
            this.applySplitRules = this.applySplitRules.bind(this);
            const allSelectors = rules
                .map(r => r.selector)
                .filter(sel => !!sel)
                .join(', ');
            if (allSelectors) {
                const shouldTraverse = (el) => !!el.querySelector(allSelectors);
                this.shouldTraverse = shouldTraverse.bind(this);
            }
            else {
                this.shouldTraverse = () => false;
            }
        }
        applySplitRules(original, clone) {
            original.classList.add(classes.toNext);
            clone.classList.add(classes.fromPrev);
            this.didSplitRules
                .filter(r => original.matches(r.selector))
                .forEach(rule => {
                rule.didSplit(original, clone);
            });
        }
        // Rules for pages
        applyPageDoneRules(pg, book) {
            this.pageRules.forEach(rule => rule.eachPage(pg, book));
        }
        finishEveryPage(book) {
            this.pageRules.forEach(rule => book.pages.forEach(pg => rule.eachPage(pg, book)));
        }
        // Rules for elements
        applyBeforeAddRules(element, book, continueOnNewPage, makeNewPage) {
            let addedElement = element;
            const matchingRules = this.beforeAddRules.filter(rule => addedElement.matches(rule.selector));
            matchingRules.forEach(rule => {
                addedElement = rule.beforeAdd(addedElement, book, continueOnNewPage, makeNewPage);
            });
            return addedElement;
        }
        applyAfterAddRules(originalElement, book, continueOnNewPage, makeNewPage) {
            let addedElement = originalElement;
            const attemptRecovery = (el) => recoverFromRule(el, book, continueOnNewPage);
            const matchingRules = this.afterAddRules.filter(rule => addedElement.matches(rule.selector));
            const uniqueRules = dedupe(matchingRules);
            uniqueRules.forEach(rule => {
                const retry = (el) => {
                    attemptRecovery(el);
                    return rule.afterAdd(el, book, continueOnNewPage, makeNewPage, () => giveUp(rule, el));
                };
                addedElement = rule.afterAdd(addedElement, book, continueOnNewPage, makeNewPage, retry);
            });
            return addedElement;
        }
    }

    const sec = (ms) => (ms / 1000).toFixed(2);
    const estimateFor = (content) => {
        const start = window.performance.now();
        const capacity = content.querySelectorAll('*').length;
        let timeWaiting = 0;
        let completed = 0;
        return {
            increment: () => {
                completed += 1;
            },
            addWaitTime: (t) => {
                timeWaiting += t;
            },
            get progress() {
                return completed / capacity;
            },
            end: () => {
                const end = window.performance.now();
                const total = end - start;
                const layout = total - timeWaiting;
                console.log(`📖 Layout ready in ${sec(layout)}s (plus ${sec(timeWaiting)}s waiting for images)`);
            },
        };
    };

    const makeBook = async (content, rules, updateProgress) => {
        if (!Page.isSizeValid())
            throw Error('Page is too small');
        const estimator = estimateFor(content);
        const ruleSet = new RuleSet(rules);
        const book = new Book();
        const pageNumberOffset = ruleSet.pageNumberOffset;
        const makeNewPage = () => new Page();
        const finishPage = (page, allowOverflow) => {
            // finished with this page, can display
            book.updatePageOrder();
            annotatePages(book.pages, pageNumberOffset);
            ruleSet.applyPageDoneRules(page, book);
            page.validateEnd(allowOverflow);
            book.validate();
        };
        const addPageToBook = (allowOverflow = false) => {
            const oldPage = book.currentPage;
            if (oldPage)
                finishPage(oldPage, allowOverflow);
            const newPage = makeNewPage();
            book.currentPage = newPage;
            book.addPage(newPage);
            updateProgress(book, estimator.progress);
            newPage.validate();
            return newPage;
        };
        const makeNextRegion = () => {
            const newPage = addPageToBook();
            return newPage.flow;
        };
        const applySplit = ruleSet.applySplitRules;
        const dontSplitSel = ruleSet.selectorsNotToSplit;
        const canSplit = (element) => {
            if (dontSplitSel.some(sel => element.matches(sel))) {
                return false;
            }
            if (element.parentElement)
                return canSplit(element.parentElement);
            return true;
        };
        const beforeAdd = (elementToAdd, continueInNextRegion) => {
            ruleSet.applyBeforeAddRules(elementToAdd, book, continueInNextRegion, makeNewPage);
        };
        const afterAdd = (addedElement, continueInNextRegion) => {
            estimator.increment();
            return ruleSet.applyAfterAddRules(addedElement, book, continueInNextRegion, makeNewPage);
        };
        // init
        content.style.margin = '0';
        content.style.padding = '0';
        await flowIntoRegions({
            content,
            createRegion: makeNextRegion,
            applySplit,
            canSplit,
            beforeAdd,
            afterAdd,
            shouldTraverse: ruleSet.shouldTraverse,
            didWaitFor: t => estimator.addWaitTime(t),
        });
        book.updatePageOrder();
        annotatePages(book.pages, pageNumberOffset);
        ruleSet.finishEveryPage(book);
        estimator.end();
        return book;
    };

    const fetchContent = async (url, selector) => {
        const response = await fetch(url);
        if (response.status !== 200) {
            throw Error(`Response ${response.status}: Could not load file at "${url}"`);
        }
        const fetchedContent = await response.text();
        const el = parseHTML(fetchedContent, selector);
        if (!(el instanceof HTMLElement)) {
            throw Error(`Could not find element that matches selector "${selector}" in response from ${url}`);
        }
        return el;
    };
    const getContentAsElement = async (content) => {
        if (content instanceof HTMLElement)
            return content;
        if (typeof content === 'string') {
            const el = document.querySelector(content);
            if (!(el instanceof HTMLElement)) {
                throw Error(`Could not find element that matches selector "${content}"`);
            }
            return el;
        }
        if (typeof content === 'object' && content.url) {
            return fetchContent(content.url, content.selector);
        }
        throw Error('Content source must be an element or selector');
    };

    // TODO: This is not a particularly robust check.
    const supportsCustomSheetSize = !!window.hasOwnProperty('chrome');
    const getSheetSizeLabels = (pageSize) => {
        const sizeName = `${pageSize.width} × ${pageSize.height}`;
        if (!supportsCustomSheetSize) {
            return [
                [SheetSize.LETTER_PORTRAIT, 'Default Page Size *'],
                [
                    SheetSize.LETTER_PORTRAIT,
                    "Only Chrome supports custom page sizes. Set in your browser's print dialog instead.",
                ],
            ];
        }
        return [
            [SheetSize.AUTO, `${sizeName}`],
            [SheetSize.AUTO_BLEED, `${sizeName} + Bleed`],
            [SheetSize.AUTO_MARKS, `${sizeName} + Marks`],
            [SheetSize.LETTER_PORTRAIT, 'Letter Portrait'],
            [SheetSize.LETTER_LANDSCAPE, 'Letter Landscape'],
            [SheetSize.A4_PORTRAIT, 'A4 Portrait'],
            [SheetSize.A4_LANDSCAPE, 'A4 Landscape'],
        ];
    };
    const marksLabels = [
        [SheetMarks.NONE, 'No Marks'],
        [SheetMarks.CROP, 'Crop Marks'],
        [SheetMarks.BLEED, 'Bleed Marks'],
        [SheetMarks.BOTH, 'Crop and Bleed'],
    ];
    const modeLabels = [
        [ViewerMode.PREVIEW, 'Grid'],
        [ViewerMode.FLIPBOOK, 'Flipbook'],
        [ViewerMode.PRINT, 'Print Preview'],
    ];
    const layoutLabels = [
        [SheetLayout.PAGES, '1 Page / Sheet'],
        [SheetLayout.SPREADS, '1 Spread / Sheet'],
        [SheetLayout.BOOKLET, 'Booklet Sheets'],
    ];

    const row = (cls, ...children) => {
        return div(`${cls}.row`, ...children);
    };
    // Button
    const btn = (cls, attrs, label) => {
        return button(`.control.btn${cls}`, attrs, label);
    };
    const dropdown = (attrs, options) => {
        const selectVal = div('.select-val', 'Value');
        const selectEl = select('.select', attrs, ...options);
        selectVal.textContent = selectEl.options[selectEl.selectedIndex].text;
        return div('.select-wrap.control', selectVal, selectEl);
    };
    const enumDropdown = (entries, initialValue, changeHandler) => {
        const eventHandler = (e) => {
            const rawVal = e.target.value;
            const chosenEntry = entries.filter(entry => entry[0].toString() === rawVal)[0];
            if (chosenEntry) {
                changeHandler(chosenEntry[0]);
            }
            else {
                throw Error('Selected unknown value');
            }
        };
        return dropdown({ onchange: eventHandler }, entries.map(entry => {
            const el = option({ value: entry[0] }, entry[1]);
            if (entry[0] === initialValue) {
                el.selected = true;
            }
            return el;
        }));
    };

    const renderPrintOptions = (state, actions) => {
        const shouldShowMarks = state.paper !== SheetSize.AUTO && state.paper !== SheetSize.AUTO_BLEED;
        const sizeLabels = getSheetSizeLabels(state.pageSize);
        return row('.print-options', row(null, enumDropdown(layoutLabels, state.layout, actions.setLayout)), row(null, enumDropdown(sizeLabels, state.paper, actions.setPaper)), shouldShowMarks
            ? row(null, enumDropdown(marksLabels, state.marks, actions.setMarks))
            : '');
    };
    class Controls {
        constructor() {
            this.element = div('.controls');
        }
        update(state, actions) {
            const oldElement = this.element;
            const newElement = this.render(state, actions);
            oldElement.replaceWith(newElement);
            this.element = newElement;
        }
        render(state, actions) {
            const print = () => {
                actions.setMode(ViewerMode.PRINT);
                setTimeout(window.print, 10);
            };
            const shouldShowPrint = state.mode === ViewerMode.PRINT;
            return div('.controls', row('.view-row', enumDropdown(modeLabels, state.mode, actions.setMode)), shouldShowPrint ? renderPrintOptions(state, actions) : '', btn('.btn-print.btn-main', { onclick: print }, 'Print'));
        }
    }

    const padPages = (pages, makePage) => {
        if (pages.length % 2 !== 0) {
            const pg = makePage();
            pages.push(pg);
        }
        const spacerPage = makePage();
        const spacerPage2 = makePage();
        spacerPage.element.style.visibility = 'hidden';
        spacerPage2.element.style.visibility = 'hidden';
        pages.unshift(spacerPage);
        pages.push(spacerPage2);
        return pages;
    };

    const twoPageSpread = (...children) => {
        return div('.spread-wrapper.spread-centered.spread-size', ...children);
    };
    const onePageSpread = (...children) => {
        return div('.spread-wrapper.spread-centered.page-size', ...children);
    };
    const renderGridViewer = (bookPages, isTwoUp) => {
        const pages = isTwoUp ? padPages(bookPages, () => new Page()) : bookPages;
        const gridLayout = document.createDocumentFragment();
        if (isTwoUp) {
            for (let i = 0; i < pages.length; i += 2) {
                const wrap = twoPageSpread(pages[i].element, pages[i + 1].element);
                gridLayout.appendChild(wrap);
            }
        }
        else {
            pages.forEach(pg => {
                const wrap = onePageSpread(pg.element);
                gridLayout.appendChild(wrap);
            });
        }
        return gridLayout;
    };

    const directions = ['top', 'bottom', 'left', 'right'];
    const bleedMarks = () => directions.map(dir => div(`.mark-bleed-${dir}`));
    const cropMarks = () => directions.map(dir => div(`.mark-crop-${dir}`));
    const pageSheetMarks = () => div('.page-size.print-mark-wrap', ...cropMarks(), ...bleedMarks());
    const spreadSheetMarks = () => div('.spread-size.print-mark-wrap', div('.mark-crop-fold'), ...cropMarks(), ...bleedMarks());
    const bookletMeta = (i, len) => {
        const isFront = i % 4 === 0;
        const sheetIndex = Math.round((i + 1) / 4) + 1;
        return div('.print-meta', `Sheet ${sheetIndex} of ${len / 4}: ${isFront ? 'Outside' : 'Inside'}`);
    };

    const orderPagesBooklet = (pages, makePage) => {
        while (pages.length % 4 !== 0) {
            const spacerPage = makePage();
            spacerPage.element.style.visibility = 'hidden';
            pages.push(spacerPage);
        }
        const bookletOrder = [];
        const len = pages.length;
        for (let i = 0; i < len / 2; i += 2) {
            bookletOrder.push(pages[len - 1 - i]);
            bookletOrder.push(pages[i]);
            bookletOrder.push(pages[i + 1]);
            bookletOrder.push(pages[len - 2 - i]);
        }
        return bookletOrder;
    };

    const twoPageSpread$1 = (...children) => {
        return div('.spread-wrapper', ...children);
    };
    const onePageSpread$1 = (...children) => {
        return div('.spread-wrapper', ...children);
    };
    const renderSheetViewer = (bookPages, _doubleSided, layout) => {
        const isTwoUp = layout !== SheetLayout.PAGES;
        const isSpreads = layout === SheetLayout.SPREADS;
        const isBooklet = layout === SheetLayout.BOOKLET;
        let pages = bookPages;
        if (isSpreads)
            pages = padPages(pages, () => new Page());
        else if (isBooklet)
            pages = orderPagesBooklet(pages, () => new Page());
        const printLayout = document.createDocumentFragment();
        const marks = isTwoUp ? spreadSheetMarks : pageSheetMarks;
        const spread = isTwoUp ? twoPageSpread$1 : onePageSpread$1;
        const printSheet = (...children) => {
            return div('.print-sheet', spread(...children));
        };
        if (isTwoUp) {
            for (let i = 0; i < pages.length; i += 2) {
                const spreadMarks = marks();
                if (isBooklet) {
                    const meta = bookletMeta(i, pages.length);
                    spreadMarks.appendChild(meta);
                }
                const sheet = printSheet(div('.page-bleed-clip.page-bleed-clip-left', pages[i].element), div('.page-bleed-clip.page-bleed-clip-right', pages[i + 1].element), spreadMarks);
                sheet.classList.add(classes.sheetSpread);
                printLayout.appendChild(sheet);
            }
        }
        else {
            pages.forEach(pg => {
                const sheet = printSheet(pg.element, marks());
                sheet.classList.add(pg.isLeft ? classes.sheetLeft : classes.sheetRight);
                printLayout.appendChild(sheet);
            });
        }
        return printLayout;
    };

    const renderFlipbookViewer = (bookPages, doubleSided) => {
        const pages = padPages(bookPages, () => new Page());
        const flipLayout = document.createDocumentFragment();
        const sizer = div('.spread-size.flip-sizer');
        const flapHolder = div('.spread-size.flap-holder');
        sizer.appendChild(flapHolder);
        flipLayout.appendChild(sizer);
        const flaps = [];
        let currentLeaf = -1;
        let leftOffset = 4;
        if (pages.length * leftOffset > 60) {
            leftOffset = 60 / pages.length;
        }
        flapHolder.style.width = `${pages.length * leftOffset}px`;
        const setLeaf = (n) => {
            let newLeaf = n;
            if (newLeaf === currentLeaf)
                newLeaf += 1;
            currentLeaf = newLeaf;
            let zScale = 4;
            if (flaps.length * zScale > 200)
                zScale = 200 / flaps.length;
            flaps.forEach((flap, i, arr) => {
                // + 0.5 so left and right are even
                const z = (arr.length - Math.abs(i - newLeaf + 0.5)) * zScale;
                flap.style.transform = `translate3d(${i < newLeaf ? 4 : 0}px,0,${z}px) rotateY(${i < newLeaf ? -180 : 0}deg)`;
            });
        };
        let leafIndex = 0;
        for (let i = 1; i < pages.length - 1; i += doubleSided ? 2 : 1) {
            leafIndex += 1;
            const li = leafIndex;
            const flap = div('.page3d');
            flap.addEventListener('click', () => {
                const newLeaf = li - 1;
                setLeaf(newLeaf);
            });
            const rightPage = pages[i].element;
            let leftPage;
            rightPage.classList.add(prefixer('page3d-front'));
            flap.appendChild(rightPage);
            if (doubleSided) {
                flap.classList.add(prefixer('doubleSided'));
                leftPage = pages[i + 1].element;
            }
            else {
                leftPage = div('.page');
            }
            leftPage.classList.add(prefixer('page3d-back'));
            flap.appendChild(leftPage);
            // TODO: Dynamically add/remove pages.
            // Putting 1000s of elements onscreen
            // locks up the browser.
            flap.style.left = `${i * leftOffset}px`;
            flaps.push(flap);
            flapHolder.appendChild(flap);
        }
        setLeaf(0);
        return flipLayout;
    };

    const getScrollAsPercent = () => {
        if (!document || !document.scrollingElement)
            return 0;
        const el = document.scrollingElement;
        return el.scrollTop / el.scrollHeight;
    };
    const scrollToPercent = (newVal) => {
        if (!document.scrollingElement)
            return;
        const el = document.scrollingElement;
        el.scrollTop = el.scrollHeight * newVal;
    };
    const scrollToBottom = () => {
        const scroll = document.scrollingElement;
        if (!scroll)
            return;
        const scrollMax = scroll.scrollHeight - scroll.offsetHeight;
        scroll.scrollTop = scrollMax;
    };

    /* global BINDERY_VERSION */
    function errorView (title, text) {
        return div('.error', div('.error-title', title), div('.error-text', text), div('.error-footer', `Bindery ${BINDERY_VERSION}`));
    }

    const isCommandP = (e) => {
        return (e.ctrlKey || e.metaKey) && e.keyCode === 80;
    };
    // Automatically switch into print mode
    const listenForPrint = (beforePrint) => {
        if (window.matchMedia) {
            const mediaQueryList = window.matchMedia('print');
            mediaQueryList.addListener(mql => {
                if (mql.matches) {
                    // before print
                    beforePrint();
                }
            });
        }
        document.body.addEventListener('keydown', e => {
            if (isCommandP(e)) {
                e.preventDefault();
                beforePrint();
                setTimeout(() => window.print(), 200);
            }
        });
    };

    const throttleProgressBar = throttleFrame();
    const throttleRender = throttleTime(100);
    const throttleResize = throttleTime(50);
    const document$1 = window.document;
    const pageSpread = (...pgs) => {
        return div('.spread-wrapper.spread-centered.spread-size', ...pgs);
    };
    class Viewer {
        constructor({ pageSetup, mode, layout, marks }) {
            this.hasRendered = false;
            this.pageSetup = pageSetup;
            this.controls = new Controls();
            this.updateControls();
            this.progressBar = div('.progress-bar');
            this.content = div('.zoom-content');
            this.scaler = div('.zoom-scaler', this.content);
            this.element = div('.root', this.progressBar, this.controls.element, this.scaler);
            this.isDoubleSided = true;
            this.sheetLayout = layout;
            this.marks = marks;
            this.mode = mode;
            this.element.classList.add(classes.viewPreview);
            this.currentLeaf = 0;
            listenForPrint(() => {
                this.mode = ViewerMode.PRINT;
                this.render();
            });
            window.addEventListener('resize', () => {
                throttleResize(() => this.scaleToFit());
            });
            this.setInProgress(true);
            this.setMarks(marks);
            this.show();
        }
        updateControls() {
            this.controls.update({
                // Initial props
                paper: this.pageSetup.paper,
                layout: this.sheetLayout,
                mode: this.mode,
                marks: this.marks,
                pageSize: this.pageSetup.displaySize,
            }, {
                // Actions
                setMode: this.setMode.bind(this),
                setPaper: this.setSheetSize.bind(this),
                setLayout: this.setLayout.bind(this),
                setMarks: this.setMarks.bind(this),
            });
        }
        setMode(newMode) {
            if (newMode === this.mode)
                return;
            this.mode = newMode;
            this.updateControls();
            this.render();
        }
        get isInProgress() {
            return this.element.classList.contains(classes.inProgress);
        }
        setInProgress(newVal) {
            this.element.classList.toggle(classes.inProgress, newVal);
        }
        get isTwoUp() {
            return this.sheetLayout !== SheetLayout.PAGES;
        }
        setShowingCropMarks(newVal) {
            this.element.classList.toggle(classes.showCrop, newVal);
        }
        setShowingBleedMarks(newVal) {
            this.element.classList.toggle(classes.showBleedMarks, newVal);
        }
        setShowingBleed(newVal) {
            this.element.classList.toggle(classes.showBleed, newVal);
        }
        get isViewing() {
            return document$1.body.classList.contains(classes.isViewing);
        }
        set isViewing(newVal) {
            document$1.body.classList.toggle(classes.isViewing, newVal);
        }
        setSheetSize(newVal) {
            this.pageSetup.paper = newVal;
            this.pageSetup.updateStyleVars();
            this.mode = ViewerMode.PRINT;
            this.render();
            this.scaleToFit();
            setTimeout(() => {
                this.scaleToFit();
            }, 300);
        }
        setLayout(newVal) {
            if (newVal === this.sheetLayout)
                return;
            this.sheetLayout = newVal;
            this.pageSetup.printTwoUp = this.isTwoUp;
            this.pageSetup.updateStyleVars();
            this.mode = ViewerMode.PRINT;
            this.render();
        }
        setMarks(newVal) {
            this.marks = newVal;
            this.updateControls();
            this.setShowingCropMarks(newVal === SheetMarks.CROP || newVal === SheetMarks.BOTH);
            this.setShowingBleedMarks(newVal === SheetMarks.BLEED || newVal === SheetMarks.BOTH);
        }
        displayError(title, text) {
            this.show();
            if (!this.error) {
                this.error = errorView(title, text);
                this.element.appendChild(this.error);
                scrollToBottom();
                if (this.book) {
                    const flow = this.book.currentPage.flow;
                    if (flow)
                        flow.currentElement.style.outline = '3px solid red';
                }
            }
        }
        clear() {
            this.book = undefined;
            this.lastSpreadInProgress = undefined; // TODO: Make this clearer, after first render
            this.content.innerHTML = '';
        }
        show() {
            if (this.element.parentNode)
                return;
            document$1.body.appendChild(this.element);
            this.isViewing = true;
        }
        hide() {
            var _a;
            // TODO this doesn't work if the target is an existing node
            (_a = this.element.parentNode) === null || _a === void 0 ? void 0 : _a.removeChild(this.element);
            this.isViewing = false;
        }
        render(newBook) {
            if (newBook)
                this.book = newBook;
            if (!this.book)
                return;
            this.show();
            this.updateControls();
            this.element.classList.remove(...allModeClasses);
            this.element.classList.add(classForMode(this.mode));
            this.setShowingBleed(this.mode === ViewerMode.PRINT);
            const prevScroll = getScrollAsPercent();
            this.setProgressAmount(1);
            window.requestAnimationFrame(() => {
                if (!this.book)
                    throw Error('Book missing');
                const pages = this.book.pages.slice();
                const render = this.renderFunctionFor(this.mode);
                const fragment = render(pages, this.isDoubleSided, this.sheetLayout);
                this.content.innerHTML = '';
                this.content.append(fragment);
                if (!this.hasRendered)
                    this.hasRendered = true;
                else
                    scrollToPercent(prevScroll);
                this.scaleToFit();
            });
        }
        renderFunctionFor(mode) {
            if (mode === ViewerMode.PREVIEW)
                return renderGridViewer;
            else if (mode === ViewerMode.FLIPBOOK)
                return renderFlipbookViewer;
            else if (mode === ViewerMode.PRINT)
                return renderSheetViewer;
            throw Error(`Invalid layout mode: ${this.mode} (type ${typeof this.mode})`);
        }
        setProgressAmount(newVal) {
            if (newVal < 1) {
                throttleProgressBar(() => {
                    this.progressBar.style.transform = `scaleX(${newVal})`;
                });
            }
            else {
                this.progressBar.style.transform = '';
            }
        }
        updateProgress(book, estimatedProgress) {
            this.book = book;
            this.setProgressAmount(estimatedProgress);
            if (!document$1.scrollingElement)
                return;
            const scroller = document$1.scrollingElement;
            // don't bother rerendering if preview is out of view
            const scrollTop = scroller.scrollTop;
            const scrollH = scroller.scrollHeight;
            const h = scroller.offsetHeight;
            if (scrollH > h * 3 && scrollTop < h)
                return;
            // don't rerender too often
            throttleRender(() => this.renderProgress(book, estimatedProgress));
        }
        renderProgress(book, estimatedProgress) {
            const needsZoomUpdate = !this.content.firstElementChild;
            const sideBySide = this.mode === ViewerMode.PREVIEW ||
                (this.mode === ViewerMode.PRINT &&
                    this.sheetLayout !== SheetLayout.PAGES);
            const limit = sideBySide ? 2 : 1;
            book.pages.forEach((page, i) => {
                if (this.content.contains(page.element) &&
                    page.element.parentNode !== this.content)
                    return;
                if (this.lastSpreadInProgress &&
                    this.lastSpreadInProgress.children.length < limit) {
                    this.lastSpreadInProgress.append(page.element);
                    return;
                }
                this.lastSpreadInProgress = pageSpread(page.element);
                if (i === 0 && sideBySide) {
                    const spacer = new Page();
                    spacer.element.style.visibility = 'hidden';
                    this.lastSpreadInProgress.insertBefore(spacer.element, this.lastSpreadInProgress.firstElementChild);
                }
                this.content.append(this.lastSpreadInProgress);
            });
            if (needsZoomUpdate)
                this.scaleToFit();
        }
        scaleToFit() {
            if (!this.content.firstElementChild)
                return;
            const prevScroll = getScrollAsPercent();
            this.scaler.style.transform = `scale(${this.scaleThatFits})`;
            scrollToPercent(prevScroll);
        }
        get scaleThatFits() {
            const viewerW = this.scaler.getBoundingClientRect().width;
            const contentW = this.content.getBoundingClientRect().width;
            return Math.min(1, viewerW / contentW);
        }
    }

    /* global BINDERY_VERSION */
    const vals = (obj) => {
        return Object.keys(obj).map(k => obj[k]);
    };
    const nextFrame = () => new Promise(resolve => {
        requestAnimationFrame(t => resolve(t));
    });
    class Bindery {
        constructor(opts = {}) {
            var _a, _b, _c, _d, _e, _f, _g, _h, _j;
            console.log(`📖 Bindery ${BINDERY_VERSION}`);
            validateRuntimeOptions(opts, {
                name: 'makeBook',
                autorun: RuntimeTypes.bool,
                content: RuntimeTypes.any,
                view: RuntimeTypes.enum(...vals(ViewerMode)),
                pageNumberOffset: RuntimeTypes.number,
                pageSetup: RuntimeTypes.shape({
                    name: 'pageSetup',
                    margin: RuntimeTypes.margin,
                    size: RuntimeTypes.size,
                }),
                printSetup: RuntimeTypes.shape({
                    name: 'printSetup',
                    bleed: RuntimeTypes.length,
                    layout: RuntimeTypes.enum(...vals(SheetLayout)),
                    marks: RuntimeTypes.enum(...vals(SheetMarks)),
                    paper: RuntimeTypes.enum(...vals(SheetSize)),
                }),
                rules: RuntimeTypes.array,
            });
            this.autorun = (_a = opts.autorun) !== null && _a !== void 0 ? _a : true;
            this.autoupdate = (_b = opts.autoupdate) !== null && _b !== void 0 ? _b : false;
            this.pageSetup = new PageSetup(opts.pageSetup, opts.printSetup);
            const startLayout = (_d = (_c = opts.printSetup) === null || _c === void 0 ? void 0 : _c.layout) !== null && _d !== void 0 ? _d : SheetLayout.PAGES;
            const startMarks = (_f = (_e = opts.printSetup) === null || _e === void 0 ? void 0 : _e.marks) !== null && _f !== void 0 ? _f : SheetMarks.CROP;
            this.viewer = new Viewer({
                pageSetup: this.pageSetup,
                mode: (_g = opts.view) !== null && _g !== void 0 ? _g : ViewerMode.PREVIEW,
                marks: startMarks,
                layout: startLayout,
            });
            if (!opts.content) {
                this.viewer.displayError('Content not specified', 'You must include a source element, selector, or url');
                throw Error('Bindery: You must include a source element or selector');
            }
            if (opts.ControlsComponent) {
                this.viewer.displayError('Controls are now included', 'Please remove the controls component');
                throw Error('Bindery: controls are now included');
            }
            this.rules = attributeRules;
            this.rules.push(new Rule({ pageNumberOffset: (_h = opts.pageNumberOffset) !== null && _h !== void 0 ? _h : 0 }));
            (_j = opts.rules) === null || _j === void 0 ? void 0 : _j.forEach(rule => {
                if (rule instanceof rules.Rule) {
                    this.rules.push(rule);
                }
                else {
                    throw Error(`Bindery: The following is not an instance of Bindery.Rule and will be ignored: ${rule}`);
                }
            });
            if (this.autorun)
                this.makeBook(opts.content);
        }
        // Convenience constructor
        static makeBook(opts = {}) {
            var _a;
            opts.autorun = (_a = opts.autorun) !== null && _a !== void 0 ? _a : true;
            return new Bindery(opts);
        }
        cancel() {
            this.viewer.hide();
            if (this.content)
                this.content.style.display = '';
        }
        async makeBook(contentDescription) {
            try {
                this.content = await getContentAsElement(contentDescription);
            }
            catch (e) {
                this.viewer.show();
                this.viewer.displayError('', e.message);
                // throw e;
                return undefined;
            }
            this.content.style.display = '';
            const content = this.content.cloneNode(true);
            this.content.style.display = 'none';
            this.viewer.clear(); // In case we're updating an existing layout
            this.viewer.show();
            this.pageSetup.updateStyleVars();
            this.viewer.setInProgress(true);
            const onProgress = (currentBook, progress) => {
                this.viewer.updateProgress(currentBook, progress);
            };
            try {
                const book = await makeBook(content, this.rules, onProgress);
                this.viewer.setProgressAmount(1);
                await nextFrame();
                this.viewer.render(book);
                this.viewer.setInProgress(false);
                return book;
            }
            catch (e) {
                this.viewer.setInProgress(false);
                this.viewer.displayError("Layout couldn't complete", e.message);
                // throw e;
                return undefined;
            }
        }
    }

    ___$insertStyle(".📖-page {\n  width: var(--bindery-page-width);\n  height: var(--bindery-page-height);\n  position: relative;\n  display: flex;\n  flex-direction: column;\n  flex-wrap: nowrap;\n}\n\n.📖-continuation {\n  text-indent: unset !important;\n}\n\nli.continuation {\n  list-style: none !important;\n}\n\n.📖-flow-box {\n  position: relative;\n  margin-top: var(--bindery-margin-top);\n  flex: 1 1 auto;\n  min-height: 0;\n}\n\n.📖-footer {\n  margin-top: 8pt;\n  margin-bottom: var(--bindery-margin-bottom);\n  flex: 0 1 auto;\n  z-index: 2;\n}\n\n.📖-flow-box,\n.📖-footer {\n  margin-left: var(--bindery-margin-inner);\n  margin-right: var(--bindery-margin-outer);\n}\n\n.📖-left .📖-flow-box,\n.📖-left .📖-footer {\n  margin-left: var(--bindery-margin-outer);\n  margin-right: var(--bindery-margin-inner);\n}\n\n.📖-page-background {\n  position: absolute;\n  z-index: 0;\n  overflow: hidden;\n  top: calc(-1 * var(--bindery-bleed));\n  bottom: calc(-1 * var(--bindery-bleed));\n  left: calc(-1 * var(--bindery-bleed));\n  right: calc(-1 * var(--bindery-bleed));\n}\n.📖-left > .📖-page-background {\n  right: 0;\n}\n.📖-right > .📖-page-background {\n  left: 0;\n}\n\n.📖-sup {\n  font-size: 0.667em;\n}\n\n.📖-running-header,\n.📖-footer {\n  font-size: 10pt;\n}\n\n.📖-running-header {\n  position: absolute;\n  text-align: center;\n  top: 0.25in;\n}\n.📖-left .📖-running-header {\n  text-align: left;\n  left: var(--bindery-margin-outer);\n}\n.📖-right .📖-running-header {\n  right: var(--bindery-margin-outer);\n  text-align: right;\n}\n\n.📖-page-size-rotated {\n  height: var(--bindery-page-width);\n  width: var(--bindery-page-height);\n}\n\n.📖-spread-size {\n  height: var(--bindery-page-height);\n  width: calc(var(--bindery-page-width) * 2);\n}\n\n.📖-spread-size-rotated {\n  width: var(--bindery-page-height);\n  height: calc(var(--bindery-page-width) * 2);\n}\n\n.📖-spread.📖-right > .📖-page-background {\n  left: calc(-100% - var(--bindery-bleed));\n}\n\n.📖-spread.📖-left > .📖-page-background {\n  right: calc(-100% - var(--bindery-bleed));\n}\n\n@media screen {\n  .📖-viewing .📖-controls {\n    display: flex !important;\n  }\n}\n.📖-controls {\n  font: 14px/1.4 -apple-system, BlinkMacSystemFont, \"Roboto\", sans-serif;\n  display: none;\n  flex-direction: row;\n  align-items: start;\n  position: sticky;\n  top: 0;\n  left: 0;\n  right: 0;\n  z-index: 99;\n  margin: auto;\n  color: var(--bindery-ui-text);\n  padding: 8px;\n  overflow: hidden;\n  -webkit-font-smoothing: antialiased;\n}\n.📖-controls * {\n  font: inherit;\n  color: inherit;\n  margin: 0;\n  padding: 0;\n  box-sizing: border-box;\n}\n.📖-controls a {\n  color: var(--bindery-ui-accent);\n  text-decoration: none;\n}\n\n.📖-row {\n  position: relative;\n  display: flex;\n  flex-wrap: wrap;\n  align-items: start;\n  cursor: default;\n  user-select: none;\n}\n\n.📖-controls .📖-btn {\n  -webkit-appearance: none;\n  cursor: pointer;\n  display: inline-block;\n  margin-right: 8px;\n  text-decoration: none;\n}\n.📖-controls .📖-btn:hover {\n  background: rgba(0, 0, 0, 0.04);\n}\n.📖-controls .📖-btn:active {\n  background: rgba(0, 0, 0, 0.08);\n}\n.📖-controls .📖-btn:last-child {\n  margin-right: 0;\n}\n\n.📖-control {\n  border-radius: 6px;\n  color: var(--bindery-ui-text);\n  padding: 4px 8px;\n  border: 1px solid #ddd;\n  margin-right: 12px;\n}\n\n.📖-controls .📖-btn-main {\n  position: absolute;\n  top: 8px;\n  right: 8px;\n  background: var(--bindery-ui-accent);\n  border-color: var(--bindery-ui-accent);\n  color: white;\n}\n.📖-controls .📖-btn-main:hover {\n  background: var(--bindery-ui-accent);\n  opacity: 0.7;\n}\n.📖-controls .📖-btn-main:active {\n  background: black;\n  opacity: 1;\n}\n\n.📖-view-row {\n  transition: all 0.3s;\n}\n.📖-in-progress .📖-view-row {\n  opacity: 0;\n  pointer-events: none;\n}\n.📖-debug .📖-view-row {\n  display: none;\n}\n\n.📖-btn-print {\n  margin-left: auto;\n  transition: all 0.3s;\n}\n.📖-in-progress .📖-btn-print {\n  opacity: 0;\n  pointer-events: none;\n}\n\n.📖-controls .📖-select-wrap {\n  padding-right: 24px;\n  transition: all 0.2s;\n  white-space: nowrap;\n  width: 100%;\n  position: relative;\n}\n.📖-controls .📖-select-wrap:after {\n  content: \"\";\n  position: absolute;\n  right: 9px;\n  top: 12px;\n  border-bottom: 1px solid;\n  border-right: 1px solid;\n  padding: 0px;\n  border: 4px solid transparent;\n  border-top-color: currentColor;\n}\n.📖-controls .📖-select-wrap:hover {\n  background-color: rgba(0, 0, 0, 0.04);\n}\n.📖-controls .📖-select-wrap:active {\n  background-color: rgba(0, 0, 0, 0.04);\n}\n\n.📖-select {\n  cursor: pointer;\n  position: absolute;\n  top: 0;\n  left: 0;\n  opacity: 0;\n  -webkit-appearance: none;\n  -moz-appearance: none;\n  padding: 4px 8px;\n  color: black;\n  border: transparent;\n  width: 100%;\n  font-size: 18px;\n}\n\n.📖-view-print .📖-controls {\n  background: var(--bindery-ui-bg);\n}\n\n@media screen and (max-width: 960px) {\n  .📖-in-progress .📖-controls {\n    background: transparent;\n    box-shadow: none;\n  }\n\n  .📖-controls {\n    background: var(--bindery-ui-bg);\n    flex-direction: column;\n  }\n}\n@media screen and (max-width: 500px) {\n  .📖-view-print .📖-controls {\n    background: var(--bindery-ui-bg);\n  }\n\n  .📖-view-row {\n    margin-bottom: 8px;\n  }\n\n  .📖-print-options {\n    flex-direction: column;\n    align-items: stretch;\n    width: 100%;\n  }\n  .📖-print-options .📖-row + .📖-row {\n    margin-top: 8px;\n  }\n  .📖-print-options .📖-select-wrap {\n    margin: 0;\n  }\n}\n.📖-rotate-container.📖-rotate-clockwise,\n.📖-left .📖-rotate-container.📖-rotate-spread-clockwise,\n.📖-right .📖-rotate-container.📖-rotate-inward,\n.📖-left .📖-rotate-container.📖-rotate-outward {\n  transform: rotate(90deg) translate3d(0, -100%, 0);\n  transform-origin: top left;\n}\n\n.📖-rotate-container.📖-rotate-counterclockwise,\n.📖-left .📖-rotate-container.📖-rotate-spread-counterclockwise,\n.📖-left .📖-rotate-container.📖-rotate-inward,\n.📖-right .📖-rotate-container.📖-rotate-outward {\n  transform: rotate(-90deg) translate3d(-100%, 0, 0);\n  transform-origin: top left;\n}\n\n.📖-rotate-container {\n  position: absolute;\n}\n.📖-left .📖-rotate-container.📖-rotate-clockwise .📖-page-background {\n  top: 0;\n}\n.📖-right .📖-rotate-container.📖-rotate-clockwise .📖-page-background {\n  bottom: 0;\n}\n.📖-left .📖-rotate-container.📖-rotate-counterclockwise .📖-page-background {\n  bottom: 0;\n}\n.📖-right .📖-rotate-container.📖-rotate-counterclockwise .📖-page-background {\n  top: 0;\n}\n.📖-rotate-container.📖-rotate-inward .📖-page-background {\n  bottom: 0;\n}\n.📖-rotate-container.📖-rotate-outward .📖-page-background {\n  top: 0;\n}\n.📖-right .📖-rotate-container.📖-rotate-spread-clockwise {\n  transform: rotate(90deg) translate3d(0, -50%, 0);\n  transform-origin: top left;\n}\n.📖-right .📖-rotate-container.📖-rotate-spread-counterclockwise {\n  transform: rotate(-90deg) translate3d(-100%, -50%, 0);\n  transform-origin: top left;\n}\n\n:root {\n  --bindery-ui-bg: #f4f4f4;\n  --bindery-ui-accent: black;\n  --bindery-ui-text: black;\n}\n\n@media screen {\n  .📖-root {\n    transition: opacity 0.2s;\n    opacity: 1;\n    background: var(--bindery-ui-bg);\n    z-index: 99;\n    position: relative;\n    min-height: 100vh;\n  }\n\n  .📖-progress-bar {\n    transform-origin: top left;\n    transform: scaleY(0);\n    position: fixed;\n    left: 0;\n    top: 0;\n    right: 0;\n    background: var(--bindery-ui-accent);\n    transition: transform 0.2s;\n    height: 2px;\n    z-index: 99;\n  }\n  .📖-in-progress .📖-progress-bar {\n    transform: scaleX(0);\n  }\n\n  .📖-zoom-content {\n    padding: 10px;\n    background: var(--bindery-ui-bg);\n    left: 0;\n    right: 0;\n    margin: auto;\n  }\n  .📖-view-preview .📖-zoom-content {\n    min-width: calc(20px + var(--bindery-spread-width));\n  }\n  .📖-view-flip .📖-zoom-content {\n    min-width: calc(1.1 * var(--bindery-spread-width));\n  }\n  .📖-view-print .📖-zoom-content {\n    min-width: calc(20px + var(--bindery-sheet-width));\n  }\n  .📖-zoom-content > .📖-page {\n    margin: auto;\n  }\n\n  .📖-measure-area {\n    position: fixed;\n    padding: 50px 20px;\n    z-index: 99;\n    visibility: hidden;\n    top: 0;\n    left: 0;\n    width: 0;\n    height: 0;\n    overflow: hidden;\n  }\n\n  .📖-is-overflowing {\n    border-bottom: 1px solid magenta;\n  }\n\n  .📖-print-sheet {\n    margin: 0 auto;\n  }\n\n  .📖-error {\n    font: 16px/1.4 -apple-system, BlinkMacSystemFont, \"Roboto\", sans-serif;\n    padding: 15vh 15vw;\n    z-index: 999;\n    position: fixed;\n    top: 0;\n    left: 0;\n    right: 0;\n    bottom: 0;\n    background: rgba(244, 244, 244, 0.7);\n  }\n  .📖-error-title {\n    font-size: 1.5em;\n    margin-bottom: 16px;\n  }\n  .📖-error-text {\n    margin-bottom: 16px;\n    white-space: pre-line;\n  }\n  .📖-error-footer {\n    opacity: 0.5;\n    font-size: 0.66em;\n    text-transform: uppercase;\n    letter-spacing: 0.02em;\n  }\n\n  .📖-show-bleed .📖-print-sheet {\n    background: white;\n    outline: 1px solid rgba(0, 0, 0, 0.1);\n    box-shadow: 0px 1px 3px rgba(0, 0, 0, 0.2);\n    margin: 20px auto;\n  }\n}\n@media screen {\n  .📖-page {\n    background: white;\n    box-shadow: 0 0 0 0.5px rgba(0, 0, 0, 0.3);\n    overflow: hidden;\n  }\n  .📖-show-bleed .📖-page {\n    box-shadow: none;\n    overflow: visible;\n  }\n  .📖-show-bleed .📖-page:after {\n    content: \"\";\n    outline: 1px dotted rgba(0, 255, 255, 0.7);\n    position: absolute;\n    left: 0;\n    right: 0;\n    bottom: 0;\n    top: 0;\n    z-index: 99;\n  }\n\n  .📖-placeholder-num {\n    visibility: hidden !important;\n  }\n}\n.📖-print-sheet {\n  width: var(--bindery-sheet-width);\n  height: var(--bindery-sheet-height);\n}\n\n.📖-page-bleed-clip {\n  overflow: hidden;\n}\n\n.📖-page-bleed-clip-left {\n  padding-left: calc(var(--bindery-bleed) + 12pt);\n}\n\n.📖-page-bleed-clip-right {\n  padding-right: calc(var(--bindery-bleed) + 12pt);\n}\n\n.📖-show-crop .📖-print-sheet .📖-page-bleed-clip,\n.📖-show-bleed-marks .📖-print-sheet .📖-page-bleed-clip {\n  padding-top: calc(var(--bindery-bleed) + 12pt);\n  padding-bottom: calc(var(--bindery-bleed) + 12pt);\n}\n\n.📖-print-sheet-spread .📖-spread-wrapper {\n  margin-left: auto;\n  margin-right: auto;\n}\n\n.📖-viewing {\n  margin: 0;\n}\n\n.📖-zoom-scaler {\n  transform-origin: top left;\n  transform-style: preserve-3d;\n}\n\n.📖-print-sheet {\n  position: relative;\n  overflow: hidden;\n  display: flex;\n  align-items: center;\n  justify-content: center;\n  transition: all 0.2s;\n}\n\n.📖-print-sheet-left {\n  justify-content: left;\n}\n\n.📖-print-sheet-right {\n  justify-content: right;\n}\n\n.📖-spread-wrapper {\n  position: relative;\n  display: flex;\n  justify-content: center;\n  margin: auto;\n}\n\n.📖-spread-centered {\n  margin: 0 auto 32px;\n}\n\n@page {\n  margin: 0;\n}\n@media print {\n  .📖-root * {\n    -webkit-print-color-adjust: exact;\n    color-adjust: exact;\n  }\n\n  .📖-controls {\n    display: none !important;\n  }\n\n  .📖-print-sheet {\n    contain: layout;\n    /* prevent margin collapse */\n    margin: 0 auto;\n    page-break-after: always;\n  }\n\n  .📖-zoom-scaler[style] {\n    transform: none !important;\n  }\n}\n.📖-print-mark-wrap {\n  display: none;\n  position: absolute;\n  pointer-events: none;\n  top: 0;\n  bottom: 0;\n  left: 0;\n  right: 0;\n  z-index: 999;\n  margin: auto;\n}\n.📖-show-crop .📖-print-mark-wrap, .📖-show-bleed-marks .📖-print-mark-wrap {\n  display: block;\n}\n.📖-show-crop .📖-print-mark-wrap > [class*=crop] {\n  display: block;\n}\n.📖-show-bleed-marks .📖-print-mark-wrap > [class*=bleed] {\n  display: block;\n}\n.📖-print-mark-wrap > div {\n  display: none;\n  position: absolute;\n  overflow: hidden;\n}\n.📖-print-mark-wrap > div::before, .📖-print-mark-wrap > div::after {\n  content: \"\";\n  display: block;\n  position: absolute;\n}\n.📖-print-mark-wrap > div:before {\n  top: 0;\n  left: 0;\n}\n.📖-print-mark-wrap > div:after {\n  bottom: 0;\n  right: 0;\n}\n\n.📖-mark-crop-fold,\n.📖-mark-crop-left,\n.📖-mark-crop-right,\n.📖-mark-bleed-left,\n.📖-mark-bleed-right {\n  width: 1px;\n  margin: auto;\n}\n.📖-mark-crop-fold::before, .📖-mark-crop-fold:after, .📖-mark-crop-left::before, .📖-mark-crop-left:after, .📖-mark-crop-right::before, .📖-mark-crop-right:after, .📖-mark-bleed-left::before, .📖-mark-bleed-left:after, .📖-mark-bleed-right::before, .📖-mark-bleed-right:after {\n  width: 1px;\n  height: var(--bindery-mark-length);\n  background-image: linear-gradient(to right, black 0%, black 51%, transparent 51%);\n  background-size: 1px 100%;\n}\n\n.📖-mark-crop-top,\n.📖-mark-crop-bottom,\n.📖-mark-bleed-top,\n.📖-mark-bleed-bottom {\n  height: 1px;\n}\n.📖-mark-crop-top::before, .📖-mark-crop-top:after, .📖-mark-crop-bottom::before, .📖-mark-crop-bottom:after, .📖-mark-bleed-top::before, .📖-mark-bleed-top:after, .📖-mark-bleed-bottom::before, .📖-mark-bleed-bottom:after {\n  width: var(--bindery-mark-length);\n  height: 1px;\n  background-image: linear-gradient(to bottom, black 0%, black 51%, transparent 51%);\n  background-size: 100% 1px;\n}\n\n.📖-mark-crop-fold {\n  right: 0;\n  left: 0;\n}\n.📖-mark-crop-left {\n  left: 0;\n}\n.📖-mark-crop-right {\n  right: 0;\n}\n.📖-mark-crop-top {\n  top: 0;\n}\n.📖-mark-crop-bottom {\n  bottom: 0;\n}\n\n.📖-print-meta {\n  padding: var(--bindery-mark-length);\n  text-align: center;\n  font-family: -apple-system, BlinkMacSystemFont, \"Roboto\", sans-serif;\n  font-size: 8pt;\n  display: block !important;\n  position: absolute;\n  bottom: -60pt;\n  left: 0;\n  right: 0;\n}\n\n.📖-mark-bleed-left,\n.📖-mark-bleed-right,\n.📖-mark-crop-left,\n.📖-mark-crop-right,\n.📖-mark-crop-fold {\n  top: calc(-1 * var(--bindery-mark-length) - var(--bindery-bleed));\n  bottom: calc(-1 * var(--bindery-mark-length) - var(--bindery-bleed));\n}\n\n.📖-mark-bleed-top,\n.📖-mark-bleed-bottom,\n.📖-mark-crop-top,\n.📖-mark-crop-bottom {\n  left: calc(-12pt - var(--bindery-bleed));\n  right: calc(-12pt - var(--bindery-bleed));\n}\n\n.📖-mark-bleed-left {\n  left: calc(-1 * var(--bindery-bleed));\n}\n\n.📖-mark-bleed-right {\n  right: calc(-1 * var(--bindery-bleed));\n}\n\n.📖-mark-bleed-top {\n  top: calc(-1 * var(--bindery-bleed));\n}\n\n.📖-mark-bleed-bottom {\n  bottom: calc(-1 * var(--bindery-bleed));\n}\n\n.📖-view-flip .📖-zoom-scaler {\n  transform-origin: center left;\n}\n\n.📖-flap-holder {\n  perspective: 5000px;\n  transform-style: preserve-3d;\n  position: absolute;\n  top: 0;\n  right: 0;\n  left: 0;\n  bottom: 0;\n  margin: auto;\n  transform-style: preserve-3d;\n}\n\n.📖-flip-sizer {\n  position: relative;\n  margin: auto;\n  padding: 0 20px;\n  box-sizing: content-box;\n  height: 90vh !important;\n}\n\n.📖-page3d {\n  margin: auto;\n  width: var(--bindery-page-width);\n  height: var(--bindery-page-height);\n  transform: rotateY(0);\n  transform-style: preserve-3d;\n  transform-origin: left;\n  transition: transform 0.5s, box-shadow 0.1s;\n  position: absolute;\n  left: 0;\n  right: 0;\n  top: 0;\n  bottom: 0;\n}\n.📖-page3d:hover {\n  box-shadow: 2px 0 4px rgba(0, 0, 0, 0.2);\n}\n.📖-page3d.📖-flipped {\n  transform: rotateY(-180deg);\n}\n.📖-page3d .📖-page {\n  position: absolute;\n  backface-visibility: hidden;\n  -webkit-backface-visibility: hidden;\n  box-shadow: 0 0 2px rgba(0, 0, 0, 0.1);\n}\n.📖-page3d .📖-page3d-front {\n  transform: rotateY(0);\n}\n.📖-page3d .📖-page3d-back {\n  transform: rotateY(-180deg);\n}");

    /* global BINDERY_VERSION */
    const constants = {
        View: ViewerMode,
        Paper: SheetSize,
        Layout: SheetLayout,
        Marks: SheetMarks,
        version: BINDERY_VERSION
    };
    const BinderyWithRules = Object.assign(Bindery, rules, constants);

    return BinderyWithRules;

})));
//# sourceMappingURL=bindery.umd.js.map
